import errno
import logging
import os
import subprocess32
import threading
import unittest

from pykit import logutil

logger = logging.getLogger(__name__)


def subproc(script, cwd=None):

    subproc = subprocess32.Popen(['sh'],
                               close_fds=True,
                               cwd=cwd,
                               stdin=subprocess32.PIPE,
                               stdout=subprocess32.PIPE,
                               stderr=subprocess32.PIPE)

    out, err = subproc.communicate(script)

    subproc.wait()

    if subproc.returncode != 0:
        print out
        print err

    return (subproc.returncode, out, err)


def read_file(fn):
    with open(fn, 'r') as f:
        return f.read()


def rm_file(fn):
    try:
        os.unlink(fn)
    except OSError as e:
        if e.errno == errno.ENOENT:
            pass
        else:
            raise


class TestFileHandler(unittest.TestCase):

    def test_concurrent_write_and_remove(self):

        l = logutil.make_logger(base_dir='/tmp',
                                log_name='rolling',
                                log_fn='rolling.out',
                                level=logging.DEBUG,
                                fmt='message')

        n = 10240
        sess = {'running': True}

        def _remove():
            while sess['running']:
                rm_file('/tmp/rolling.out')

        th = threading.Thread(target=_remove)
        th.daemon = True
        th.start()

        for ii in range(n):
            l.debug('123')

        sess['running'] = False
        th.join()


class TestLogutil(unittest.TestCase):

    def setUp(self):

        rm_file('/tmp/t.out')

        # root logger
        logutil.make_logger(base_dir='/tmp',
                            log_fn='t.out',
                            level=logging.DEBUG,
                            fmt='message')

    def test_get_root_log_fn(self):

        # instant

        code, out, err = subproc(
            'python -c "from pykit import logutil; print logutil.get_root_log_fn()"')
        self.assertEqual(0, code)
        self.assertEqual('__instant_command__.out', out.strip())

        code, out, err = subproc(
            'echo "from pykit import logutil; print logutil.get_root_log_fn()" | python')
        self.assertEqual(0, code)
        self.assertEqual('__stdin__.out', out.strip())

        # load by file

        code, out, err = subproc(
            'python foo.py', cwd=os.path.dirname(__file__))
        self.assertEqual(0, code)
        self.assertEqual('foo.out', out.strip())

    def test_deprecate(self):

        fmt = '{fn}::{ln} in {func}\n  {statement}'
        logutil.deprecate('foo', fmt=fmt, sep='\n')

        cont = read_file('/tmp/t.out')

        self.assertRegexpMatches(
            cont,
            '^Deprecated: foo')
        self.assertRegexpMatches(
            cont,
            'test_logutil.py::\d+ in test_deprecate\n  logutil.deprecate')

    def test_stack_list(self):

        stack = logutil.stack_list()
        last = stack[-1]

        self.assertEqual('test_logutil.py', os.path.basename(last[0]))
        self.assertTrue(isinstance(last[1], int))
        self.assertEqual('test_stack_list', last[2])
        self.assertRegexpMatches(last[3], '^ *stack = ')

    def test_format_stack(self):

        cases = (
            ([('0', 1, 2, 3)], '0-1-2-3'),
            ([('0', 1, 2, 3),
              ('a', 'b', 'c', 'd')], '0-1-2-3\na-b-c-d'),
        )

        for inp, expected in cases:

            rst = logutil.stack_format(
                inp, fmt='{fn}-{ln}-{func}-{statement}', sep='\n')
            self.assertEqual(expected, rst)

    def test_stack_str(self):

        rst = logutil.stack_str(fmt='{fn}-{ln}-{func}-{statement}', sep=' ')
        self.assertRegexpMatches(
            rst,
            ' test_logutil.py-\d+-test_stack_str- *rst = ')

    def test_get_datefmt(self):

        cases = (
                (None,      None),
                ('default', None),
                ('time',    '%H:%M:%S'),
                ('%H%M%S',  '%H%M%S'),
        )

        for inp, expected in cases:
            rst = logutil.get_datefmt(inp)

            self.assertEqual(expected, rst)

    def test_get_fmt(self):

        cases = (
                (None,
                 '[%(asctime)s,%(process)d-%(thread)d,%(filename)s,%(lineno)d,%(levelname)s] %(message)s'),
                ('default',
                 '[%(asctime)s,%(process)d-%(thread)d,%(filename)s,%(lineno)d,%(levelname)s] %(message)s'),
                ('time_level',  "[%(asctime)s,%(levelname)s] %(message)s"),
                ('message',     '%(message)s'),
                ('%(message)s', '%(message)s'),
        )

        for inp, expected in cases:
            rst = logutil.get_fmt(inp)
            self.assertEqual(expected, rst)

    def test_make_logger(self):

        rm_file('/tmp/tt')

        l = logutil.make_logger(base_dir='/tmp',
                                log_name='m',
                                log_fn='tt',
                                level='INFO',
                                fmt='%(message)s',
                                datefmt='%H%M%S'
                                )

        l.debug('debug')
        l.info('info')

        cont = read_file('/tmp/tt').strip()

        self.assertEqual(cont, 'info')

    def test_make_logger_with_config(self):

        code, out, err = subproc(
            'python make_logger_with_config.py', cwd=os.path.dirname(__file__))
        self.assertEqual(0, code)
        self.assertEqual(out.strip(), 'info')

    def test_make_formatter(self):
        # how to test logging.Formatter?
        pass

    def test_make_file_handler(self):

        rm_file('/tmp/handler_change')

        l = logutil.make_logger(base_dir='/tmp',
                                log_name='h',
                                log_fn='dd',
                                level='INFO',
                                fmt='%(message)s',
                                datefmt='%H%M%S'
                                )
        l.handlers = []
        handler = logutil.make_file_handler(base_dir='/tmp',
                                            log_fn='handler_change',
                                            fmt='%(message)s',
                                            datefmt='%H%M%S')
        l.addHandler(handler)

        l.debug('debug')
        l.info('info')

        cont = read_file('/tmp/handler_change').strip()

        self.assertEqual(cont, 'info')

    def test_make_file_handler_with_config(self):

        code, out, err = subproc(
            'python make_file_handler_with_config.py', cwd=os.path.dirname(__file__))
        self.assertEqual(0, code)
        self.assertEqual(out.strip(), 'info')

    def test_add_std_handler(self):
        rm_file('/tmp/stdlog')

        code, out, err = subproc(
            'python stdlog.py', cwd=os.path.dirname(__file__))
        self.assertEqual(0, code)
        self.assertEqual('error', out.strip())

    def test_set_logger_level(self):

        cases = (
            (None,                      'debug1\ndebug2'),
            ('1_prefix',                'debug1\ndebug2\ndebug2'),
            (('1_prefix', '2_prefix'),  'debug1\ndebug2'),
            (('not_exist',),            'debug1\ndebug2\ndebug1\ndebug2'),
            (('not_exist', '1_prefix'), 'debug1\ndebug2\ndebug2'),
        )

        for inp, expected in cases:
            rm_file('/tmp/ss')

            logger1 = logutil.make_logger(base_dir='/tmp',
                                          log_name='1_prefix_1',
                                          log_fn='ss',
                                          level='DEBUG',
                                          fmt='%(message)s',
                                          datefmt='%H%M%S')

            logger2 = logutil.make_logger(base_dir='/tmp',
                                          log_name='2_prefix_1',
                                          log_fn='ss',
                                          level='DEBUG',
                                          fmt='%(message)s',
                                          datefmt='%H%M%S')
            logger1.debug('debug1')
            logger2.debug('debug2')

            logutil.set_logger_level(level='INFO', name_prefixes=inp)

            logger1.debug('debug1')
            logger2.debug('debug2')

            content = read_file('/tmp/ss')

            self.assertEqual(expected, content.strip())



import logging

import psutil

from pykit import fsutil

logger = logging.getLogger(__name__)


def get_pid_from_file(pid_file):
    data = fsutil.read_file(pid_file)
    return int(data)


def get_all_pids(pid):
    all_pids = []

    try:
        process = psutil.Process(pid)

    except psutil.NoSuchProcess as e:
        logger.info('process %d does not exist' % pid)
        return all_pids

    except Exception as e:
        logger.exception('faild to get process of pid: %d, %s' %
                         (pid, repr(e)))
        return all_pids

    all_pids.append(process.pid)

    children = process.children(recursive=True)
    for process in children:
        all_pids.append(process.pid)

    return all_pids




# index is started from 1.
#
#          1
#     2          3
#   4    5     6    7
#                            level
#          1                   0
#    10         11             1
# 100  101   110  111          2
#
# - On each level, widths of level in binary number are the same and just equals level+1.
# - The most significant bit are always 1.
# - The 2nd significant bit indicate it is on left(0) or right(1) level-1 sub-heap.
# - The 3rd significant bit indicate it is on left(0) or right(1) level-2 sub-heap.
# - ...

primitive_types = (type(None), int, long, bool, basestring, tuple, float, complex)


class HeapError(Exception):
    pass


class Duplicate(HeapError):
    pass


class Empty(HeapError):
    pass


class NotFound(HeapError):
    pass


class Node(list):

    # node in heap:
    # node[0] is left child
    # node[1] is right child

    def __init__(self, userdata):

        super(Node, self).__init__([None, None])

        self.parent = None
        self.userdata = userdata

    def min_child(self):
        if self[0] is None:
            return self[1]

        if self[1] is None:
            return self[0]

        if self[0].userdata < self[1].userdata:
            return self[0]
        else:
            return self[1]

    def __lt__(self, b):
        return self.userdata < b.userdata

    def __str__(self):

        ud_str = str(self.userdata)

        mark = '> '
        width = 16
        item_width = width + len(mark)

        ud_str = ('{0:>' + str(width) + '}').format(ud_str)
        ud_str = ud_str + mark

        if self == [None, None]:
            return ud_str

        if self[0] is None:
            sl = ['-']
        else:
            sl = str(self[0]).split('\n')
        if self[1] is None:
            sr = ['-']
        else:
            sr = str(self[1]).split('\n')

        sl = [' ' * item_width + x for x in sl]
        sr = [' ' * item_width + x for x in sr]
        sl[0] = ud_str + sl[0][item_width:]

        return '\n'.join(sl) + '\n' + '\n'.join(sr)


# wrapper of primitive type
class Primitive(object):

    def __init__(self, obj):
        self.data = obj

    def __str__(self):
        return str(self.data)

    def __lt__(self, b):
        return self.data < b.data


class RefHeap(object):

    def __init__(self, iterable=()):
        self.size = 0
        self.root = None

        # map user data `obj` to heap node instance.
        self.userdata_map = {}

        for i in iterable:
            self.push(i)

    def push(self, obj):
        # wrap primitive type with an object or no duplicate can be added

        if isinstance(obj, primitive_types):
            obj = Primitive(obj)

        obj_id = id(obj)

        if obj_id in self.userdata_map:
            raise Duplicate('object is already in heap')

        node = Node(obj)
        self.userdata_map[obj_id] = node

        self.size += 1

        if self.size == 1:
            self.root = node
            return

        pos = self.size

        p = self.node_at(_parent_index(pos))

        p[_child_lr_idx(pos)] = node
        node.parent = p

        self._sift(node)

    def get(self):
        self.assert_not_empty()
        return self.return_val(self.root)

    def pop(self):
        self.assert_not_empty()
        return self.remove_node(self.root)

    def pop_all(self, map=lambda x: x):
        rst = []
        while self.size > 0:
            rst.append(map(self.pop()))
        return rst

    def remove(self, obj):
        node = self._get_object_node(obj)
        return self.remove_node(node)

    def remove_node(self, node):

        rst = self.return_val(node)

        last = self.node_at(self.size)

        self.swap(node, last)
        self.remove_last(last)
        self._sift(node)

        if self.size == 0:
            self.root = None

        return rst

    def sift(self, obj):
        if isinstance(obj, primitive_types):
            raise ValueError('primitive type does not support sift, just replace it')

        node = self._get_object_node(obj)
        return self._sift(node)

    def _get_object_node(self, obj):
        if id(obj) not in self.userdata_map:
            raise NotFound('object is not found in heap, id:' + str(obj))

        node = self.userdata_map[id(obj)]
        return node

    def _sift(self, node):

        x = node.userdata
        xid = id(x)

        # upwards
        up = False
        while (node.parent is not None
               and x < node.parent.userdata):

            # move parent down to node
            self.userdata_map[id(node.parent.userdata)] = node
            node.userdata = node.parent.userdata

            node = node.parent
            up = True

        node.userdata = x
        self.userdata_map[xid] = node

        if up:
            return

        # downwards
        while True:

            _min = node.min_child()
            if _min is None or not (_min.userdata < x):
                break

            # move _min up to node
            self.userdata_map[id(_min.userdata)] = node
            node.userdata = _min.userdata

            node = _min

        node.userdata = x
        self.userdata_map[xid] = node

    def swap(self, a, b):
        # we do not swap node, but exchange node.userdata only.
        self.userdata_map[id(a.userdata)], self.userdata_map[id(b.userdata)] = b, a
        a.userdata, b.userdata = b.userdata, a.userdata

    def remove_last(self, last):
        if last.parent is not None:
            last.parent[_child_lr_idx(self.size)] = None
        last.parent = None

        del self.userdata_map[id(last.userdata)]
        last.userdata = None

        self.size -= 1

    def return_val(self, node):
        rst = node.userdata
        if isinstance(rst, Primitive):
            rst = rst.data
        return rst

    def node_at(self, idx):

        # idx starts from 1

        p = self.root
        lvl = index_level(idx)

        for l in range(lvl - 1, -1, -1):
            lr = (idx >> l) & 1
            p = p[lr]

        return p

    def assert_not_empty(self):
        if self.size == 0:
            raise Empty()

    def __str__(self):
        return str(self.root)


def index_level(idx):

    # root is at level 0

    lvl = 0
    while idx > 1:
        idx = _parent_index(idx)
        lvl += 1

    return lvl


def _parent_index(c):
    return c // 2


def _child_lr_idx(idx):
    return idx % 2



#!/usr/bin/env python2
# coding: utf-8

from collections import namedtuple

from .block_group_id import BlockGroupID
from .block_index import BlockIndex
from .idbase import IDBase
from .server import DriveID

BlockIDLen = 48


class BlockBaseError(Exception):
    pass


class BlockIDError(BlockBaseError):
    pass


class BlockID(namedtuple('_BlockID', 'type block_group_id block_index drive_id bg_seq'), IDBase):

    _tostr_fmt = '{type}{block_group_id}{block_index}{drive_id}{bg_seq:0>10}'

    @classmethod
    def parse(cls, block_id):

        if len(block_id) != BlockIDLen:
            raise BlockIDError('Block id length should be {0}, but is {1}: {2}'.format(
                BlockIDLen, len(block_id), block_id))

        return BlockID(block_id[:2],
                       block_id[2:18],
                       block_id[18:22],
                       block_id[22:38],
                       block_id[-10:])

    def __new__(clz, type, block_group_id, block_index, drive_id, bg_seq):

        if not isinstance(block_group_id, BlockGroupID):
            block_group_id = BlockGroupID.parse(block_group_id)

        if not isinstance(block_index, BlockIndex):
            block_index = BlockIndex.parse(block_index)

        if not isinstance(drive_id, DriveID):
            drive_id = DriveID.parse(drive_id)

        bg_seq = int(bg_seq)

        return super(BlockID, clz).__new__(clz, type, block_group_id, block_index, drive_id, bg_seq)



from pykit import dictutil


class TrieNode(dict):

    def __init__(self, *args, **kwargs):
        super(TrieNode, self).__init__(*args, **kwargs)

        # Total number of items with the prefix represented by this node.
        # `None` means there might be more items with this prefix thus its
        # number can not be decided yet.
        self.n = None

        # Trie branch key, a single char
        self.char = ''

        # An outstanding node is a node that there might be more following input
        # string which has its corresponding prefix.
        self.outstanding = None

        # If this node is an outstanding node to its parent node.
        # When created, a node must be an outstanding node.
        self.is_outstanding = True

        # If this node is an end of a line of string.
        # A leaf node must be an eol node.
        # But not vice verse
        self.is_eol = False

    def __str__(self):

        children_keys = sorted(self.keys())

        if self.is_outstanding:
            mark = '*'
        else:
            mark = ' '

        if self.char != '':
            c = str(self.char) + ','
        else:
            c = ''

        if len(children_keys) > 0:
            colon = ': '
        else:
            colon = ''

        line = '{mark}{c}{n}{colon}'.format(c=c, n=(self.n or '?'), mark=mark, colon=colon)

        fields = []
        for c in children_keys:
            substr = str(self[c])
            indent = ' ' * len(line)
            substr = indent + substr.replace('\n', '\n' + indent)

            fields.append(substr)

        rst = '\n'.join(fields)
        return line + rst[len(line):]


def _trie_node(parent, char):
    n = TrieNode()
    n.char = char
    parent.outstanding = n
    return n


def make_trie(sorted_iterable, node_max_num=1):

    t = TrieNode()

    for _s in sorted_iterable:

        # find longest common prefix of _s and any seen string
        node = t
        for i, c in enumerate(_s):
            if c in node:
                node = node[c]
            else:
                break

        # `node` now is at where the longest common prefix.
        # `i` points to next char not in common prefix in `_s`.

        # Since `node` is the longest prefix and all input strings are sorted,
        # the prefix represented by the child `node.outstanding` can never be a prefix of
        # any following strings.
        _squash(node.outstanding, node_max_num)

        for c in _s[i:]:
            node[c] = _trie_node(node, c)
            node = node[c]

        node.is_eol = True

        # Only leaf node is count by 1
        node.n = 1

    _squash(t, node_max_num)

    return t


def sharding(sorted_iterable, size, accuracy=None, joiner=''.join):

    if accuracy is None:
        accuracy = size / 10

    t = make_trie(sorted_iterable, node_max_num=accuracy)

    n = 0
    prev_key = None
    rst = []

    # iterate root node.
    t = {'':t}

    for ks, node in dictutil.depth_iter(t, is_allowed=lambda ks, v: v.is_eol or len(v) == 0):

        if n >= size:

            rst.append((prev_key, n))

            # strip empty root path
            prev_key = ks[1:]
            prev_key = joiner(prev_key)
            n = 0

        if len(node) == 0:
            n += node.n
        else:
            # node.is_eol == True
            n += 1

    rst.append((prev_key, n))

    return rst


def _squash(node, node_max_num):

    # If the number of strings with prefix at `node`(children of `node`) is smaller than `node_max_num`,
    # squash all children to reduce memory cost.

    if node is None:
        return

    _squash(node.outstanding, node_max_num)

    total = node.n or 0
    for subnode in node.values():
        total += subnode.n

    if total <= node_max_num:
        for c in node.keys():
            del node[c]

    node.n = total
    node.is_outstanding = False



#!/usr/bin/env python2
# coding: utf-8

import copy
from collections import defaultdict

from .block import BlockID
from .block_index import BlockIndex
from .block_index import BlockIndexError

BlockIndexLen = 4


class BlockGroupBaseError(Exception):
    pass


class BlockNotFoundError(BlockGroupBaseError):
    pass


class BlockTypeNotSupported(BlockGroupBaseError):
    pass


class BlockTypeNotSupportReplica(BlockGroupBaseError):
    pass


class BlockGroup(dict):

    def __init__(self, block_group=None, **kwargs):

        if block_group is not None:
            self.update(copy.deepcopy(block_group))

        self.update(copy.deepcopy(kwargs))

    @classmethod
    def make(cls, block_group_id, idcs, config):

        bg = cls({
            'block_group_id': block_group_id,
            'bg_seq': 0,
            'config': copy.deepcopy(config),
            'idcs': idcs,
            'blocks': {}
        })

        nr_data, nr_parity = bg['config']['ec']['in_idc']

        for idc_idx, idc in enumerate(idcs):
            for pos in range(nr_data + nr_parity):
                b_idx = BlockIndex(idc_idx, pos)
                bg['blocks'][str(b_idx)] = bg.empty_block(b_idx)

        return bg

    def get_block_type(self, block_index):

        mp = self.get_type_map()
        bi = BlockIndex.parse(str(block_index))

        try:
            return mp[bi.i][bi.j]
        except IndexError:
            # TODO refine error
            raise BlockTypeNotSupported('type x{1~n} is not supported')

    def get_type_map(self):

        cnf = self['config']['ec']

        nr_data, nr_parity = cnf['in_idc']
        nr_in_idc, nr_xor_idc = cnf['cross_idc']
        data_replica = cnf['data_replica']

        rst = []

        for i in range(nr_in_idc):

            o = ['d0'] * nr_data + ['dp'] * nr_parity

            for ii in range(data_replica - 1):
                o += ['d%d' % (ii+1)] * nr_data

            rst.append(o)

        for i in range(nr_xor_idc):
            o = ['x0'] * nr_data + ['xp'] * nr_parity
            rst.append(o)

        return rst

    def empty_block(self, block_index):

        blk_type = self.get_block_type(block_index)

        block = {
            'block_id': None,
            'size': 0,
            'type': blk_type,
            'range': [None, None],
            'is_del': 1
        }

        return block

    def mark_delete_block(self, block_index=None, block_id=None):
        block_index, block = self.get_block(block_index=block_index,
                                            block_id=block_id,
                                            raise_error=True)
        block['is_del'] = 1

    def replace_block(self, new_block, block_index=None, block_id=None):
        block_index, block = self.get_block(block_index=block_index,
                                            block_id=block_id,
                                            raise_error=True)

        self['blocks'][str(block_index)] = new_block

    def get_free_block_indexes(self, block_type=None):

        free_block_index = defaultdict(list)

        for b_idx, block in self['blocks'].items():

            if block_type is not None:
                if block['type'] != block_type:
                    continue

            if block['block_id'] is None:
                idc = self.get_block_idc(block_index=b_idx)
                free_block_index[idc].append(b_idx)

        for idc in free_block_index.keys():
            free_block_index[idc].sort()

        return free_block_index

    def get_block(self, block_index=None, block_id=None, raise_error=False):

        block = None

        if block_index is not None:
            msg = ('block_index:{i} '
                   'not found in block_group:{block_group_id}').format(i=block_index,
                                                                       **self)
            block_index = BlockIndex.parse(str(block_index))
            block = self['blocks'].get(str(block_index))

        elif block_id is not None:
            msg = ('block_id:{i} '
                   'not found in block_group:{block_group_id}').format(i=block_id,
                                                                       **self)

            block_index = BlockID.parse(block_id).block_index
            block = self['blocks'].get(str(block_index))

            if block is not None and block['block_id'] != block_id:
                block = None

        if raise_error is True and block is None:
            raise BlockNotFoundError(msg)

        return block_index, block

    def get_block_idc(self, block_index=None, block_id=None):

        blk_idx, _ = self.get_block(block_index=block_index,
                                    block_id=block_id,
                                    raise_error=True)

        return self['idcs'][blk_idx.i]

    def get_replica_block_index(self, block_index=None, block_id=None):

        if block_id is not None:
            bi, _ = self.get_block(block_id=block_id, raise_error=True)
        else:
            bi = BlockIndex.parse(str(block_index))

        blk_type = self.get_block_type(bi)

        if blk_type.endswith('p') or blk_type.startswith('x'):
            raise BlockTypeNotSupportReplica(('block type {0} '
                                              'do not support replica').format(blk_type))

        replica_block_index = []

        for rbi in self.get_replica_indexes(bi):
            if rbi.j == bi.j:
                continue

            replica_block_index.append(str(rbi))

        return replica_block_index

    def get_primary_index(self, block_index):

        nr_data, nr_parity = self['config']['ec']['in_idc']

        bi = BlockIndex.parse(str(block_index))

        j = bi.j
        if j >= nr_data:
            j -= nr_parity
            j %= nr_data

        return BlockIndex(bi.i, j)

    def get_replica_indexes(self, block_index):

        nr_data, nr_parity = self['config']['ec']['in_idc']
        data_replica = self['config']['ec']['data_replica']

        bi = self.get_primary_index(block_index)

        rst = [bi]

        for j in range(1, data_replica):
            rst.append(BlockIndex(bi.i,
                                  bi.j + nr_parity + j * nr_data))

        return rst



import errno
import os

from pykit import logutil


def read_file(fn):
    with open(fn, 'r') as f:
        return f.read()


def rm_file(fn):
    try:
        os.unlink(fn)
    except OSError as e:
        if e.errno == errno.ENOENT:
            pass
        else:
            raise


rm_file('/tmp/handler_change')

l = logutil.make_logger(base_dir='/tmp',
                        log_name='h',
                        log_fn='dd',
                        level='INFO',
                        fmt='%(message)s',
                        datefmt='%H%M%S'
                        )
l.handlers = []
handler = logutil.make_file_handler(log_fn='handler_change',
                                    fmt='%(message)s',
                                    datefmt='%H%M%S')
l.addHandler(handler)

l.debug('debug')
l.info('info')

cont = read_file('/tmp/handler_change').strip()

print cont



#!/usr/bin/env python2
# coding: utf-8

import copy
import operator
import unittest

from pykit import dictutil
from pykit import ututil

dd = ututil.dd


class TestDictDeepIter(unittest.TestCase):

    def test_depth_iter_default(self):

        cases = (
            ({}, []),
            ({'k1': 'v1'},
             [(['k1'], 'v1')]
             ),
            ({'k1': 'v1', 'k2': 'v2'},
             [
                (['k1'], 'v1'),
                (['k2'], 'v2'),
            ]),
            ({'k1': {'k11': 'v11'},
              'k2': 'v2',
              'empty': {},
              },
             ([
                 (['k1', 'k11'], 'v11'),
                 (['k2'], 'v2'),
             ]),
             )
        )

        for _in, _out in cases:

            idx = 0

            for rst in dictutil.depth_iter(_in):
                self.assertEqual(
                    _out[idx],
                    rst,
                    ('input: {_in}, output: {rst}, expected: {_out}').format(
                        _in=repr(_in),
                        _out=repr(rst),
                        rst=repr(_out[idx])
                    )
                )

                idx = idx + 1

    def test_depth_iter_ks(self):

        idx = 0
        ks = ['mykey']

        _in = {'k1': {'k11': 'v11'}, 'k2': 'v2'}
        _out = [
            (['mykey', 'k1', 'k11'], 'v11'),
            (['mykey', 'k2'], 'v2'),
        ]
        _mes = 'test argument ks in dictutil.dict_depth_iter()'

        for rst in dictutil.depth_iter(_in, ks=ks):
            self.assertEqual(
                _out[idx],
                rst,
                ('input: {_in}, output: {rst}, expected: {_out},'
                 'message: {_mes}').format(
                     _in=repr(_in),
                     _out=repr(rst),
                     rst=repr(_out[idx]),
                     _mes=_mes
                )
            )

            idx = idx + 1

    def test_depth_iter_maxdepth(self):

        idx = 0

        _in = {
            'k1': {
                'k11': {
                    'k111': {
                        'k1111': 'v1111'
                    },
                    'empty': {},
                },
            },
        }

        _out = [
            (['k1'], {'k11': {'k111': {'k1111': 'v1111'}, 'empty': {}}}),
            (['k1', 'k11'], {'k111': {'k1111': 'v1111'}, 'empty': {}}),
            (['k1', 'k11', 'empty'], {}),
            (['k1', 'k11', 'k111'], {'k1111': 'v1111'}),
            (['k1', 'k11', 'k111', 'k1111'], 'v1111')
        ]

        for depth in range(1, 5):
            for rst in dictutil.depth_iter(_in, maxdepth=depth):
                dd('depth:', depth)
                dd('_in:', _in)
                dd('_out[idx]:', _out[idx])
                dd('rst:', rst)
                self.assertEqual(_out[idx], rst)

                idx = idx + 1

    def test_depth_iter_intermediate(self):

        _in = {
            'k1': {
                'k11': {
                    'k111': {
                        'k1111': 'v1111'
                    },
                    'empty': {},
                },
            },
        }

        _out = [
            (['k1'], {'k11': {'k111': {'k1111': 'v1111'}, 'empty': {}}}),
            (['k1', 'k11'], {'k111': {'k1111': 'v1111'}, 'empty': {}}),
            (['k1', 'k11', 'empty'], {}),
            (['k1', 'k11', 'k111'], {'k1111': 'v1111'}),
            (['k1', 'k11', 'k111', 'k1111'], 'v1111')
        ]

        idx = 0
        for rst in dictutil.depth_iter(_in, intermediate=True):
            self.assertEqual(
                _out[idx],
                rst)

            idx = idx + 1

        self.assertEqual(len(_out), idx)

    def test_depth_iter_empty_as_leaf(self):

        _in = {
            'k1': {
                'k11': {
                    'k111': {
                        'k1111': 'v1111'
                    },
                    'empty': {},
                },
            },
        }

        _out = [
            (['k1', 'k11', 'empty'], {}),
            (['k1', 'k11', 'k111', 'k1111'], 'v1111'),
        ]

        idx = 0
        for rst in dictutil.depth_iter(_in, empty_leaf=True):
            self.assertEqual(_out[idx], rst)
            idx = idx + 1

        self.assertEqual(len(_out), idx)

    def test_depth_iter_empty_as_leaf_and_intermediate(self):

        _in = {
            'k1': {
                'k11': {
                    'k111': {
                        'k1111': 'v1111'
                    },
                    'empty': {},
                },
            },
        }

        _out = [
            (['k1'], {'k11': {'k111': {'k1111': 'v1111'}, 'empty': {}}}),
            (['k1', 'k11'], {'k111': {'k1111': 'v1111'}, 'empty': {}}),
            (['k1', 'k11', 'empty'], {}),
            (['k1', 'k11', 'k111'], {'k1111': 'v1111'}),
            (['k1', 'k11', 'k111', 'k1111'], 'v1111')
        ]

        idx = 0
        for rst in dictutil.depth_iter(_in, intermediate=True, empty_leaf=True):
            self.assertEqual(_out[idx], rst)
            idx = idx + 1

        self.assertEqual(len(_out), idx)

    def test_depth_iter_is_allowed(self):

        _in = {
            'k1': {
                'k11': {
                    'k111': {
                        'k1111': 'v1111'
                    },
                    'empty': {},
                },
            },
        }

        cases = (
                ({'is_allowed': lambda ks, v: isinstance(v, dict) and len(v) == 0,
                  },
                 ['k1', 'k11', 'empty'], {},
                 'choose only empty dict value',
                 ),

                ({'is_allowed': lambda ks, v: isinstance(v, str),
                  },
                 ['k1', 'k11', 'k111', 'k1111'], 'v1111',
                 'choose only string value',
                 ),

                ({'is_allowed': lambda ks, v: isinstance(v, str),
                  'intermediate': True,
                  },
                 ['k1', 'k11', 'k111', 'k1111'], 'v1111',
                 'is_allowed should override intermediate',
                 ),

                ({'is_allowed': lambda ks, v: isinstance(v, str),
                  'empty_leaf': True,
                  },
                 ['k1', 'k11', 'k111', 'k1111'], 'v1111',
                 'is_allowed should override empty_leaf',
                 ),

        )

        for kwargs, expected_ks, expected_v, msg in cases:

            dd(msg)
            dd(kwargs)
            dd('expected:', expected_ks, expected_v)

            for ks, v in dictutil.depth_iter(_in, **kwargs):
                self.assertEqual(expected_ks, ks)
                self.assertEqual(expected_v, v)


class TestDictBreadthIter(unittest.TestCase):

    def test_breadth_iter_default(self):

        cases = (
            ({}, []),
            ({'k1': 'v1'}, [(['k1'], 'v1')]),
            ({'k1': 'v1', 'k2': 'v2'}, [(['k2'], 'v2'), (['k1'], 'v1')]),
            ({'k1': {'k11': 'v11'}, 'k2': 'v2'},
             ([(['k2'], 'v2'),
               (['k1'], {'k11': 'v11'}),
               (['k1', 'k11'], 'v11')
               ])
             )
        )

        for _in, _out in cases:

            idx = 0

            for rst in dictutil.breadth_iter(_in):
                self.assertEqual(
                    _out[idx],
                    rst,
                    ('input: {_in}, output: {rst}, expected: {_out}').format(
                        _in=repr(_in),
                        _out=repr(rst),
                        rst=repr(_out[idx]),
                    )
                )

                idx = idx + 1


class TestGetter(unittest.TestCase):

    def test_getter_str(self):

        cases = (
            ('',
             [''],
             ('',),
             'lambda dic, vars={}: dic.get("", vars.get("_default", 0))'
             ),

            ('.',
             ['', ''],
             ('', '',),
             'lambda dic, vars={}: dic.get("", {}).get("", vars.get("_default", 0))'
             ),

            ('...',
             ['', '', '', ''],
             ('', '', '', ''),
             'lambda dic, vars={}: dic.get("", {}).get("", {}).get("", {}).get("", vars.get("_default", 0))'
             ),

            ('.x',
             ['', 'x'],
             ('', 'x'),
             'lambda dic, vars={}: dic.get("", {}).get("x", vars.get("_default", 0))'
             ),

            ('x..',
             ['x', '', ''],
             ('x', '', ''),
             'lambda dic, vars={}: dic.get("x", {}).get("", {}).get("", vars.get("_default", 0))'
             ),

            ('x',
             ['x'],
             ('x',),
             'lambda dic, vars={}: dic.get("x", vars.get("_default", 0))'
             ),

            ('x.y',
             ['x', 'y'],
             ('x', 'y',),
             'lambda dic, vars={}: dic.get("x", {}).get("y", vars.get("_default", 0))'
             ),

            ('x.y.zz',
             ['x', 'y', 'zz'],
             ('x', 'y', 'zz',),
             'lambda dic, vars={}: dic.get("x", {}).get("y", {}).get("zz", vars.get("_default", 0))'
             ),

            # dynamic
            ('$',
             ['$'],
             ('$',),
             'lambda dic, vars={}: dic.get(str(vars.get("", "_")), vars.get("_default", 0))'
             ),

            ('$xx',
             ['$xx'],
             ('$xx',),
             'lambda dic, vars={}: dic.get(str(vars.get("xx", "_")), vars.get("_default", 0))'
             ),

            ('$xx.$yy',
             ['$xx', '$yy'],
             ('$xx', '$yy',),
             'lambda dic, vars={}: dic.get(str(vars.get("xx", "_")), {}).get(str(vars.get("yy", "_")), vars.get("_default", 0))'
             ),

            (None,
             [1, '1', 1.0, ()],
             (1, '1', 1.0, ()),
             'lambda dic, vars={}: dic.get(1, {}).get("1", {}).get(1.0, {}).get((), vars.get("_default", 0))',
             ),

            (None,
             [('x',), ('y',)],
             (('x',), ('y',)),
             'lambda dic, vars={}: dic.get(("x",), {}).get(("y",), vars.get("_default", 0))',
             ),

            (None,
             [(1,), ('1', (1.0,),)],
             ((1,), ('1', (1.0,),)),
             'lambda dic, vars={}: dic.get((1,), {}).get(("1",(1.0,),), vars.get("_default", 0))',
             ),

            (None,
             [(1,), ('$xx', (1.0,), '$yy'), ],
             ((1,), ('$xx', (1.0,), '$yy'), ),
             'lambda dic, vars={}: dic.get((1,), {}).get((str(vars.get("xx", "_")),(1.0,),str(vars.get("yy", "_")),), vars.get("_default", 0))',
             ),

            (None,
             [{1: 1}, [1, '1']],
             ({1: 1}, [1, '1']),
             'lambda dic, vars={}: dic.get({1: 1}, {}).get([1, \'1\'], vars.get("_default", 0))')
        )

        for _in_str, _in_list, _in_tuple, _out in cases:

            self.assert_make_getter_str(_in_str, _out)

            self.assert_make_getter_str(_in_list, _out)

            self.assert_make_getter_str(_in_tuple, _out)

    def assert_make_getter_str(self, _in, _out, default=0):
        if _in is None:
            return
        rst = dictutil.make_getter_str(_in, default=default)
        self.assertEqual(_out, rst,
                         'input: {_in}, expected: {_out}, actual: {rst}'.format(
                             _in=repr(_in),
                             _out=repr(_out),
                             rst=repr(rst),
                         )
                         )

    def test_getter_str_default(self):

        cases = (

            ('x', None,
             'lambda dic, vars={}: dic.get("x", vars.get("_default", None))'
             ),

            ('x', 1,
             'lambda dic, vars={}: dic.get("x", vars.get("_default", 1))'
             ),

            ('x', True,
             'lambda dic, vars={}: dic.get("x", vars.get("_default", True))'
             ),

            ('x', "abc",
             'lambda dic, vars={}: dic.get("x", vars.get("_default", \'abc\'))'
             ),

        )

        for _in_str, _default, _out in cases:

            _in_list = [_in_str]
            _in_tuple = (_in_str,)

            self.assert_make_getter_str(_in_str, _out, default=_default)

            self.assert_make_getter_str(_in_list, _out, default=_default)

            self.assert_make_getter_str(_in_tuple, _out, default=_default)

    def test_getter_vars(self):

        cases = (

            ('', [''], ('',),
             0,
             {"x": 1, "": 1}, {},
             1
             ),

            ('.', ['', ''], ('', '',),
             0,
             {'': {'': 1}}, {},
             1
             ),

            ('...', ['', '', '', ''], ('', '', '', ''),
             0,
             {'': {'': {'': {'': 1}}}}, {},
             1
             ),

            ('.x', ['', 'x'], ('', 'x',),
             0,
             {'': {'x': 1}}, {},
             1
             ),

            ('x..', ['x', '', ''], ('x', '', '',),
             0,
             {'x': {'': {'': 1}}}, {},
             1
             ),

            ('x', ['x'], ('x',), 0,
             {}, {},
             0
             ),

            ('x', ['x'], ('x',), 0,
             {'x': {'y': 3}}, {},
             {'y': 3}
             ),

            ('x', ['x'], ('x',), 55,
             {}, {},
             55
             ),

            ('x', ['x'], ('x',), 55,
             {}, {'_default': 66},
             66
             ),

            ('x', ['x'], ('x',), 55,
             {"x": 1}, {},
             1
             ),

            ('x.y', ['x', 'y'], ('x', 'y',), 55,
             {"x": {"y": 3}}, {},
             3
             ),

            ('x.z', ['x', 'z'], ('x', 'z',), 55,
             {"x": {"y": 3}}, {},
             55
             ),

            ('x.$v', ['x', '$v'], ('x', '$v',), 55,
             {"x": {"y": 3}}, {},
             55
             ),

            ('x.$var_name', ['x', '$var_name'], ('x', '$var_name',), 55,
             {"x": {"y": 3}}, {"var_name": "y"},
             3
             ),

            ('x.$var_name.z', ['x', '$var_name', 'z'], ('x', '$var_name', 'z',), 55,
             {"x": {"y": {}}}, {"var_name": "y"},
             55
             ),

            ('x.$var_name.z', ['x', '$var_name', 'z'], ('x', '$var_name', 'z',), 55,
             {"x": {"y": {"z": 4}}}, {"var_name": "y"},
             4
             ),

            (None, [1, '1', 1.0, ()], (1, '1', 1.0, ()), 55,
             {1: {'1': {1.0: {(): 3}}}}, {},
             3
             ),

            (None, [(1,), ('$xx', (1.0,), '$yy'), ], ((1,), ('$xx', (1.0,), '$yy'),), 55,
             {(1,): {('xx', (1.0,), 'yy'): 3}}, {"xx": "xx", "yy": "yy"},
             3
             ),
        )

        for _in_str, _in_str_list, _in_str_tuple, _default, _dic, _vars, _out in cases:

            self.assert_getter_and_maker(_in_str, _dic, _out, _vars, _default)

            self.assert_getter_and_maker(_in_str_list, _dic, _out, _vars, _default)

            self.assert_getter_and_maker(_in_str_tuple, _dic, _out, _vars, _default)

    def test_getter_type_error(self):
        cases = (
            ([{1: 1}, [1, '1']],
             {},),
            (({1: 1}, [1, '1']),
             {}),
        )

        for _in, _dic in cases:
            acc = dictutil.make_getter(_in, default=0)
            with self.assertRaises(TypeError):
                acc(_dic)
            with self.assertRaises(TypeError):
                dictutil.get(_dic, _in)

    def assert_getter_and_maker(self, _in, _dic, _out, _vars, _default):

        if _in is None:
            return
        acc = dictutil.make_getter(_in, default=_default)
        rst = acc(_dic, vars=_vars)
        self.assertEqual(_out, rst,
                         'input: {_in}, {_default}, {_dic}, {_vars} expected: {_out}, actual: {rst}'.format(
                             _in=repr(_in),
                             _default=repr(_default),
                             _dic=repr(_dic),
                             _vars=repr(_vars),
                             _out=repr(_out),
                             rst=repr(rst),
                         )
                         )
        # test dictutil.get() with the same set of cases
        rst = dictutil.get(_dic, _in, vars=_vars, default=_default)
        dd('dictutil.get({dic}, {key_path} vars={vars}, default={default})'.format(
            dic=_dic,
            key_path=_in,
            vars=_vars,
            default=_default,
        ))
        dd(rst)
        self.assertEqual(_out, rst,
                         'input: {_in}, {_default}, {_dic}, {_vars} expected: {_out}, actual: {rst}'.format(
                             _in=repr(_in),
                             _default=repr(_default),
                             _dic=repr(_dic),
                             _vars=repr(_vars),
                             _out=repr(_out),
                             rst=repr(rst),
                         )
                         )

    def test_get_ignore_vars_key_error(self):

        cases = (
                ({}, '$a', {"a": "x"}, None, 0),
                ({}, '$a', {"a": "x"}, True, 0),
                ({}, ['$a', ('$a',), ], {"a": "x"}, True, 0),
        )

        for case in cases:

            dd('case: ', case)

            dic, key_path, vars, ign, expected = case
            rst = dictutil.get(dic, key_path, vars=vars, ignore_vars_key_error=ign)

            dd('rst: ', rst)

            self.assertEqual(expected, rst)

        with self.assertRaises(KeyError):
            dictutil.get({}, '$a', {}, ignore_vars_key_error=False)

        with self.assertRaises(KeyError):
            dictutil.get({}, (('1', '$a',),), {}, ignore_vars_key_error=False)

        with self.assertRaises(KeyError):
            dictutil.get({}, [('1', '$a',)], {}, ignore_vars_key_error=False)


class TestSetter(unittest.TestCase):

    def test_setter(self):

        cases = (
                ('', {},
                 10,
                 {'': 10}
                 ),

                ('.', {},
                 10,
                 {'': {'': 10}}),

                ('...', {},
                 10,
                 {'': {'': {'': {'': 10}}}}
                 ),

                ('.x', {},
                 10,
                 {'': {'x': 10}},
                 ),

                ('x..', {},
                 10,
                 {'x': {'': {'': 10}}},
                 ),

                ('a', {},
                 3,
                 {'a': 3},
                 ),

                ('a', {'exist': 44},
                 3,
                 {'a': 3, 'exist': 44},
                 ),

                ('a.b.c', {'exist': 44},
                 3,
                 {'a': {'b': {'c': 3}}, 'exist': 44},
                 ),

                ('a.b.c', {'a': {'exist': 44}},
                 3,
                 {'a': {'b': {'c': 3}, 'exist': 44}},
                 ),

                (['a', 'b', 'c'], {'a': {'exist': 44}},
                 3,
                 {'a': {'b': {'c': 3}, 'exist': 44}},
                 ),

                (('a', 'b', 'c'), {'a': {'exist': 44}},
                 3,
                 {'a': {'b': {'c': 3}, 'exist': 44}},
                 ),

                ([1, '1', 1.0, ()], {'a': {'exist': 44}},
                 3,
                 {'a': {'exist': 44}, 1: {'1': {1.0: {(): 3}}}},
                 ),

                ((1, '1', 1.0, ()), {'a': {'exist': 44}},
                 3,
                 {'a': {'exist': 44}, 1: {'1': {1.0: {(): 3}}}},
                 ),

                ([(1,), ('1', (1.0,),)], {'a': {'exist': 44}},
                 3,
                 {'a': {'exist': 44}, (1,): {('1', (1.0,),): 3}},
                 ),

                (((1,), ('1', (1.0,),)), {'a': {'exist': 44}},
                 3,
                 {'a': {'exist': 44}, (1,): {('1', (1.0,),): 3}},
                 ),

        )

        for _key_path, _dic, _value, _expect in cases:

            _set = dictutil.make_setter(_key_path)
            rst = _set(_dic, _value)

            self.assertEqual(_value, rst,
                             'input: {_key_path}, {_dic}; expected return value: {_value}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _value=repr(_value),
                                 rst=repr(rst),
                             )
                             )

            self.assertEqual(_expect, _dic,
                             'input: {_key_path}, {_dic}; expected dict: {_expect}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _expect=repr(_expect),
                                 rst=repr(rst),
                             )
                             )

    def test_setter_default(self):

        cases = (

                ('a', 'value', {},
                 {'a': 'value'},
                 ),

                ('exist', 'value', {'exist': 44},
                 {'exist': 'value'},
                 ),

                ('a.exist', 'value', {'a': {'exist': 44}},
                 {'a': {'exist': 'value'}},
                 ),

                ('a.exist', lambda *x: '_def', {'a': {'exist': 44}},
                 {'a': {'exist': '_def'}},
                 ),

                ('a.exist', lambda vars: vars.get('foo'), {'a': {'exist': 44}},
                 {'a': {'exist': 'foo'}},
                 ),

                ([1, '1', 1.0, ()], lambda vars: vars.get('foo'), {'a': {'exist': 44}},
                 {'a': {'exist': 44}, 1: {'1': {1.0: {(): 'foo'}}}},
                 ),

                ((1, '1', 1.0, ()), lambda vars: vars.get('foo'), {'a': {'exist': 44}},
                 {'a': {'exist': 44}, 1: {'1': {1.0: {(): 'foo'}}}},
                 ),

                ([(1,), ('1', (1.0,),)], lambda vars: vars.get('foo'), {'a': {'exist': 44}},
                 {'a': {'exist': 44}, (1,): {('1', (1.0,),): 'foo'}},
                 ),

                (((1,), ('1', (1.0,),)), lambda vars: vars.get('foo'), {'a': {'exist': 44}},
                 {'a': {'exist': 44}, (1,): {('1', (1.0,),): 'foo'}},
                 ),

        )

        _vars = {'foo': 'foo'}

        for _key_path, _default, _dic, _expect in cases:

            _set = dictutil.make_setter(_key_path, value=_default)
            rst = _set(_dic, vars=_vars)

            if callable(_default):
                _def = _default(_vars)
            else:
                _def = _default

            self.assertEqual(_def, rst,
                             'input: {_key_path}, {_dic}; expected return value: {_def}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _def=repr(_def),
                                 rst=repr(rst),
                             )
                             )

            self.assertEqual(_expect, _dic,
                             'input: {_key_path}, {_dic}; expected dict: {_expect}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _expect=repr(_expect),
                                 rst=repr(rst),
                             )
                             )

            rst = _set(_dic, 'specified', vars=_vars)
            self.assertEqual('specified', rst)

    def test_setter_vars_inexistent(self):

        _set = dictutil.make_setter('$a')

        with self.assertRaises(KeyError):
            _set({}, vars={})

        _set = dictutil.make_setter(['$a'])

        with self.assertRaises(KeyError):
            _set({}, vars={})

        _set = dictutil.make_setter(('$a',))

        with self.assertRaises(KeyError):
            _set({}, vars={})

    def test_setter_vars(self):

        cases = (

                ('$a', 1,  {},
                 {'aa': 1},
                 ),

                ('$a.$foo', 1,  {},
                 {'aa': {'bar': 1}},
                 ),

                ([(1,), ('$a', (1.0,), '$foo')], 1, {},
                 {(1,): {('aa', (1.0,), 'bar'): 1}}
                 ),

                (((1,), ('$a', (1.0,), '$foo'),), 1, {},
                 {(1,): {('aa', (1.0,), 'bar'): 1}}
                 ),

        )

        _vars = {'foo': 'bar',
                 'a': 'aa',
                 }

        for _key_path, _default, _dic, _expect in cases:

            _set = dictutil.make_setter(_key_path, value=_default)
            rst = _set(_dic, vars=_vars)

            if callable(_default):
                _def = _default(_vars)
            else:
                _def = _default

            self.assertEqual(_def, rst,
                             'input: {_key_path}, {_dic}; expected return value: {_def}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _def=repr(_def),
                                 rst=repr(rst),
                             )
                             )

            self.assertEqual(_expect, _dic,
                             'input: {_key_path}, {_dic}; expected dict: {_expect}, actual: {rst}'.format(
                                 _key_path=repr(_key_path),
                                 _dic=repr(_dic),
                                 _expect=repr(_expect),
                                 rst=repr(rst),
                             )
                             )

            rst = _set(_dic, 'specified', vars=_vars)
            self.assertEqual('specified', rst)

    def test_setter_incr(self):

        cases = (

                ('a',
                 ['a'],
                 ('a',),
                 1, {},
                 {'a': 1},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 1, {'a': 1},
                 {'a': 2},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 'foo', {},
                 {'a': 'foo'},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 'foo', {'a': 'bar'},
                 {'a': 'barfoo'},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 (1, 2), {},
                 {'a': (1, 2)},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 (1, 2), {'a': (0, 1)},
                 {'a': (0, 1, 1, 2)},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 [1, 2], {},
                 {'a': [1, 2]},
                 ),

                ('a',
                 ['a'],
                 ('a',),
                 [1, 2], {'a': [0, 1]},
                 {'a': [0, 1, 1, 2]},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 1, {},
                 {'aa': {'bar': 1}},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 1.1, {},
                 {'aa': {'bar': 1.1}},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 'suffix', {'aa': {'bar': 'prefix-'}},
                 {'aa': {'bar': 'prefix-suffix'}},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 ('b',), {'aa': {}},
                 {'aa': {'bar': ('b',)}},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 ('b',), {'aa': {'bar': ('a',)}},
                 {'aa': {'bar': ('a', 'b',)}},
                 ),

                ('$a.$foo',
                 ['$a', '$foo'],
                 ('$a', '$foo',),
                 ['b', ], {'aa': {'bar': ['a', ]}},
                 {'aa': {'bar': ['a', 'b', ]}},
                 ),

                (None,
                 [1, '1', 1.0, ()],
                 (1, '1', 1.0, ()),
                 11,
                 {1: {'1': {1.0: {(): 100}}}},
                 {1: {'1': {1.0: {(): 111}}}},
                 ),

                (None,
                 [(1,), ('$foo', (1.0,), '$a')],
                 ((1,), ('$foo', (1.0,), '$a'),),
                 11,
                 {(1,): {('bar', (1.0,), 'aa'): 100}},
                 {(1,): {('bar', (1.0,), 'aa'): 111}},
                 ),

        )

        _vars = {'foo': 'bar',
                 'a': 'aa',
                 }

        for _in_str, _in_list, _in_tuple, _default, _dic, _expect in cases:
            self.assert_make_setter(_default, copy.deepcopy(_dic), _expect, _in_str, _vars)
            self.assert_make_setter(_default, copy.deepcopy(_dic), _expect, _in_list, _vars)
            self.assert_make_setter(_default, copy.deepcopy(_dic), _expect, _in_tuple, _vars)

    def assert_make_setter(self, _default, _dic, _expect, _in, _vars):

        if _in is None:
            return

        dd(_in, _default, _dic, _expect)
        _set = dictutil.make_setter(_in, value=_default, incr=True)
        rst = _set(_dic, vars=_vars)
        dd('rst:', rst)
        self.assertEqual(_expect, _dic,
                         'input: {_key_path}, {_dic}; expected dict: {_expect}, actual: {rst}'.format(
                             _key_path=repr(_in),
                             _dic=repr(_dic),
                             _expect=repr(_expect),
                             rst=repr(rst),
                         )
                         )


class TestAttrDict(unittest.TestCase):

    def test_attrdict(self):

        cases = (
                ([],
                 {},
                 {}
                 ),

                ([],
                 {'a': 2},
                 {'a': 2}
                 ),

                ([{'x': 2}],
                 {'a': 4},
                 {'x': 2, 'a': 4},
                 ),
        )

        for args, kwargs, expected in cases:
            rst = dictutil.attrdict(*args, **kwargs)
            self.assertEqual(expected, rst,
                             'input: {a} {kw} {e}, rst: {rst}'.format(
                                 a=args, kw=kwargs, e=expected, rst=rst))

            for k in rst:
                self.assertEqual(rst[k], getattr(rst, k))

    def test_dict_method(self):

        ad = dictutil.attrdict(a=1, b=2)

        self.assertEqual(['a', 'b'], ad.keys())
        self.assertEqual([1, 2], ad.values())
        self.assertEqual([('a', 1), ('b', 2)], ad.items())

    def test_recursive(self):

        ad = dictutil.attrdict(
            x=1, y={'a': 3, 'b': dict(c=4), 'd': dictutil.attrdict(z=5)})

        self.assertEqual(1, ad.x)
        self.assertEqual({'a': 3, 'b': {'c': 4}, 'd': {'z': 5}}, ad.y)
        self.assertEqual(3, ad.y.a)
        self.assertEqual(4, ad.y.b.c)
        self.assertEqual(5, ad.y.d.z)

    def test_writable(self):

        ad = dictutil.attrdict(a={}, b={1: 2})
        ad['x'] = 4
        self.assertEqual(4, ad.x)
        self.assertEqual(4, ad['x'])
        ad.y = 5
        self.assertEqual(5, ad.y)
        self.assertEqual(5, ad['y'])

    def test_attrdict_copy(self):

        # reference type value are always copied when accessing

        ad = dictutil.attrdict_copy(a={}, b={'x': {'foo': 'bar'}})
        self.assertIsNot(ad.a, ad.a)
        self.assertIsNot(ad.b.x, ad.b.x)
        self.assertIsNot(ad['a'], ad['a'])
        self.assertIsNot(ad['b']['x'], ad['b']['x'])

        # value got is still a AttrDictCopy instance

        b = ad.b
        self.assertIsNot(b.x, b.x)
        self.assertIsNot(b['x'], b['x'])

        # it does not change the original value

        with self.assertRaises(KeyError):
            ad['b'] = 10

        with self.assertRaises(KeyError):
            ad.b['x'] = 10

        with self.assertRaises(KeyError):
            b['x'] = 10

        with self.assertRaises(AttributeError):
            ad.b = 10

        with self.assertRaises(AttributeError):
            ad.b.x = 10

        with self.assertRaises(AttributeError):
            ad.y = 2

    def test_attrdict_copy_as_dict(self):

        ad = dictutil.attrdict_copy(a={}, b={'x': {}})
        d = ad.as_dict()

        b2 = d['b']
        d['b']['x'] = 100
        self.assertEqual(100, b2['x'])

    def test_attr_overriding(self):

        ad = dictutil.attrdict(items=1)

        with self.assertRaises(TypeError):
            ad.items()

    def test_ref_to_same_item(self):

        x = {'a': 1}
        ad = dictutil.attrdict(u=x, v=x)

        self.assertIs(ad.u, ad.v)

        x['x'] = x
        ad = dictutil.attrdict(x)

        self.assertTrue(isinstance(ad, dictutil.AttrDict))
        self.assertTrue(isinstance(ad.x, dictutil.AttrDict))
        self.assertTrue(ad.x is not ad, 'attrdict does create a new dict')
        self.assertTrue(ad.x.x is ad.x, 'circular references work for all dict items.')
        self.assertTrue(ad.x.x.x is ad.x.x, 'circular references work for all dict items.(2)')


class TestIsSubDict(unittest.TestCase):

    def test_dict(self):

        for case in [
            (1, 1, True),
            (1, 2, False),
            ("x", "x", True),
            ("x", "b", False),
            (None, None, True),

            ({"a": 1}, {"a": 1}, True),
            ({}, {"a": None}, False),
            ({"a": 1}, {}, True),
            ({"a": 1}, None, False),
            (None, {"a": 1}, False),

            ({"a": (1, 2)}, {"a": (1, 2)}, True),
            ({"a": (1, 2, 3)}, {"a": (1, 2)}, True),

            ({"a": 1}, [], False),
            ({"a": [1, (2, 3)]}, {"a": [1, (2,)]}, True),
            ({"a": [1, (2, 3)]}, {"a": [1, (2, 4)]}, False),
            ({"a": [1, (2, 3)]}, {"a": [1, (2, 3, 4)]}, False),

            ({"a": 1, "b": 2}, {"a": 1}, True),
            ({"a": 1}, {"a": 1, "b": 2}, False),
            ({"a": 1, "b": {"c": 3, "d": 4}}, {"a": 1, "b": {"d": 4}}, True),
            ({"a": 1, "b": {"c": 3, "d": 4}}, {"a": 1, "b": 2}, False),
        ]:
            self.assertEqual(dictutil.contains(case[0], case[1]), case[2])

    def test_recursive_dict(self):
        a = {}
        a[1] = {}
        a[1][1] = a

        b = {}
        b[1] = {}
        b[1][1] = {}
        b[1][1][1] = b

        self.assertEqual(dictutil.contains(a, b), True)

    def test_recursive_dict_with_list(self):
        a = {'k': [0, 2]}
        a['k'][0] = {'k': [0, 2]}
        a['k'][0]['k'][0] = a

        b = {'k': [0, 2]}
        b['k'][0] = b

        self.assertEqual(dictutil.contains(a, b), True)


class TestAdd(unittest.TestCase):

    def test_add(self):
        cases = (
            ({},          None,       {}, {}),
            ({},          'foo',      {}, {}),
            ({},          123,        {}, {}),
            ({},          {},         {}, {}),
            ({'a': 1},    {},         {}, {'a': 1}),
            ({},          {'a': 2},   {}, {'a': 2}),
            ({'a': 1},    {'a': 2},   {}, {'a': 3}),
            ({'a': '1'},  {'a': '2'}, {}, {'a': '12'}),
            ({'a': '1'},  {'b': '2'}, {}, {'a': '1', 'b': '2'}),
            ({'a': '1'},  {'b': {}},  {}, {'a': '1', 'b': {}}),
            ({'a': {}},   {'b': {}},  {}, {'a': {}, 'b': {}}),
            ({'a': {}},   {'a': {}},  {}, {'a': {}}),

            ({'a': {'k1': 1}},
             {'a': {}},
             {},
             {'a': {'k1': 1}}),

            ({'a': {'k1': 1}},
             {'a': {'k2': 1}},
             {},
             {'a': {'k1': 1, 'k2': 1}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {},
             {'a': {'k1': 2}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'a': True}, },
             {'a': {'k1': 1}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'a': {}}, },
             {'a': {'k1': 2}}),

            ({'a': {}},
             {'a': {'k1': 1}},
             {'exclude': {'a': {}}, },
             {'a': {'k1': 1}}),

            ({'a': {}},
             {'a': {'k1': 1}},
             {'exclude': {'a': True}, },
             {'a': {}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'a': {}}, },
             {'a': {'k1': 2}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'b': True}, },
             {'a': {'k1': 2}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'b': True}, },
             {'a': {'k1': 2}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'a': {'k1': True}}, 'recursive': True},
             {'a': {'k1': 1}}),

            ({'a': {'k1': 1}},
             {'a': {'k1': 1}},
             {'exclude': {'a': {'k2': True}}, 'recursive': True},
             {'a': {'k1': 2}}),

            ({'a': {'k1': 1}, 'b': 1},
             {'a': {'k1': 1, 'k2': 1}, 'b': 1},
             {'recursive': False},
             {'a': {'k1': 1}, 'b': 2}),
        )

        for a, b, kwargs, expected in cases:
            dd('a:', a)
            dd('b:', b)
            dd('kwargs:', kwargs)
            dd('expected:', expected)

            result = dictutil.add(a, b, **kwargs)
            dd('rst:', result)

            self.assertIsNot(a, result)
            self.assertDictEqual(expected, result,
                                 repr([a, b, kwargs, result]))

            result = dictutil.addto(a, b, **kwargs)

            self.assertIs(a, result)
            self.assertDictEqual(expected, result,
                                 repr([a, b, kwargs, result]))


class TestCombine(unittest.TestCase):

    def test_combine(self):
        cases = (
            ({'a': 2},
             {'a': 3},
             operator.mul,
             None,
             {'a': 6}),

            ({'a': 2},
             {'a': 3},
             operator.mul,
             {'a': True},
             {'a': 2}),
        )

        for a, b, op, exclude, expected in cases:
            result = dictutil.combine(a, b, op, exclude=exclude)
            self.assertIsNot(a, result)
            self.assertDictEqual(expected, result,
                                 repr([a, b, op, exclude, expected, result]))

            result = dictutil.combineto(a, b, op, exclude=exclude)
            self.assertIs(a, result)
            self.assertDictEqual(expected, result,
                                 repr([a, b, op, exclude, expected, result]))


class TestDictutil(unittest.TestCase):

    def test_subdict(self):

        source_dict = {0: '0', 'a': 'a', 'none': None}

        default_dict = {'b': 'b', 'c': 'c'}

        def iter_default(k):
            return default_dict.get(k)

        def iter_keys():
            for i in (0, 'a', 'b'):
                yield i

        cases = (
            (
                ('a', 0),
                {
                    'use_default': True,
                },
                {0: '0', 'a': 'a'},
                'normal',
            ),
            (
                ('a', 'b'),
                {
                    'use_default': True,
                    'default': 0,
                },
                {'a': 'a', 'b': 0},
                'use default',
            ),
            (
                ('b', 'c'),
                {
                    'use_default': True,
                    'default': iter_default,
                },
                {'b': 'b', 'c': 'c'},
                'use default as callable',
            ),
            (
                ('b', 'c'),
                {
                    'use_default': False,
                    'default': iter_default,
                },
                {},
                'not use default',
            ),
            (
                iter_keys(),
                {
                    'use_default': True,
                },
                {0: '0', 'a': 'a', 'b': None},
                'use keys as interable',
            ),
            (
                ('none',),
                {
                    'use_default': True,
                    'default': 'not None',
                },
                {'none': None},
                'key exists and is None',
            ),
        )

        for flds, kwargs, expected, msg in cases:
            dd('msg: ', msg)
            dd('expected: ', expected)
            rst = dictutil.subdict(source_dict, flds, **kwargs)
            dd('result:   ', rst)

            self.assertEqual(expected, rst)

        test_dict = {'test': 0}
        source_dict['copy'] = test_dict
        deepcopy_cases = (
            (
                ('copy',),
                {
                    'deepcopy': True,
                },
                {'copy': {'test': 0}},
                'test deepcopy value',
            ),
            (
                ('copy',),
                {},
                {'copy': test_dict},
                'test reference value',
            ),

            (
                ('default',),
                {
                    'use_default': True,
                    'default': test_dict,
                    'deepcopy_default': True,
                },
                {'default': {'test': 0}},
                'test deepcopy default',
            ),
            (
                ('default',),
                {
                    'use_default': True,
                    'default': test_dict,
                },
                {'default': test_dict},
                'test reference default',
            ),
            (
                ('copy', 'default'),
                {
                    'use_default': True,
                    'default': test_dict,
                    'deepcopy': True,
                },
                {
                    'copy': {'test': 0},
                    'default': test_dict,
                },
                'test deepcopy value and reference default',
            ),
            (
                ('copy', 'default'),
                {
                    'use_default': True,
                    'default': test_dict,
                    'deepcopy_default': True,
                },
                {
                    'copy': test_dict,
                    'default': {'test': 0},
                },
                'test reference value and deepcopy default',
            ),
        )

        for flds, kwargs, expected, msg in deepcopy_cases:
            dd('msg :', msg)
            dd('expected: ', expected)

            rst = dictutil.subdict(source_dict, flds, **kwargs)

            dd('result:   ', rst)

            self.assertEqual(expected, rst)

            if 'deepcopy' not in kwargs:
                for k in [x for x in flds if x in source_dict]:
                    self.assertIs(expected[k], rst[k])

            if 'deepcopy_default' not in kwargs:
                for k in [x for x in flds if x not in source_dict]:
                    self.assertIs(expected[k], rst[k])



#!/usr/bin/env python
# coding: utf-8

from collections import namedtuple

from .idbase import IDBase


class BlockIndexError(ValueError):
    pass


class BlockIndex(namedtuple('_BlockIndex', 'i j'), IDBase):

    _tostr_fmt = '{i:0>2}{j:0>2}'

    @classmethod
    def parse(clz, block_index):
        if len(block_index) != 4:
            raise BlockIndexError('block_index len must be 4, but: {bi}'.format(
                bi=block_index))

        ii, jj = block_index[:2], block_index[2:]

        return BlockIndex(ii, jj)

    def __new__(clz, i, j):
        try:
            i = int(i)
            j = int(j)
        except ValueError:
            raise BlockIndexError('index i and j must be int or int-str but: {i} {j}'.format(i=i, j=j))

        if not (0 <= i <= 99):
            raise BlockIndexError('idc index must be 0~99, but: {i}'.format(i=i))
        if not (0 <= j <= 99):
            raise BlockIndexError('cross idc index must be 0~99, but: {j}'.format(j=j))

        return super(BlockIndex, clz).__new__(clz, i, j)



#!/usr/bin/env python2
# coding: utf-8

import logging
import time

logger = logging.getLogger(__name__)


def run(job):

    data = job.data

    data['result'] = data.get('echo')
    time.sleep(data.get('sleep_time', 10))



#!/usr/bin/env python2.6
# coding: utf-8

import unittest

from pykit import awssign

EMPTY_PAYLOAD_HASH = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'


class TestSigner(unittest.TestCase):
    access_key = ''
    secret_key = ''
    signer = awssign.Signer(access_key, secret_key)

    def test_uri_encode_args(self):
        args = {
            'foo': True,
            'foo/': ['bar /', '~bar', True, 'bar', True],
            '?': '&=+',
            ' ': ' '
        }
        expect = {
            'foo': True,
            'foo%2F': ['bar%20%2F', '~bar', True, 'bar', True],
            '%3F': '%26%3D%2B',
            '%20': '%20'
        }

        actual = self.signer._uri_encode_args(args)
        self.assertEqual(expect, actual)

    def test_build_canonical_query_string(self):
        test_cases = (
            ({}, '', 'case 1'),
            ({'foo': True}, 'foo=', 'case 2'),
            ({'foo': [True]}, 'foo=', 'case 3'),
            ({'foo': ['']}, 'foo=', 'case 4'),
            ({'foo': [True, 'bar', True]}, 'foo=', 'case 5'),
            ({'foo': ['bar', True], 'bar': ''}, 'bar=&foo=bar', 'case 6'),
            ({'foo': '', 'bar': True}, 'bar=&foo=', 'case 7'),
            ({'bar': [True, 'bar']}, 'bar=', 'case 8'),
            ({'X-Amz-Signature': 'foo', 'bar': 'bar'}, 'bar=bar', 'case 9'),

        )
        for args, expected, des in test_cases:
            actual = self.signer._build_canonical_query_string(args)
            self.assertEqual(expected, actual, des)

    def test_make_sha256(self):
        str = 'foo'
        expected = '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae'
        actual = self.signer._make_sha256(str)
        self.assertEqual(expected, actual)

    def test_make_hmac_sha256(self):
        key = 'foo'
        msg = 'bar'
        expected = 'f9320baf0249169e73850cd6156ded0106e2bb6ad8cab01b7bbbebe6d1065317'
        actual = self.signer._make_hmac_sha256(key, msg, True)
        self.assertEqual(expected, actual)

    def test_query_string_to_args(self):
        test_cases = (
            ('', {}),
            ('foo%2F', {'foo/': True}),
            ('foo%2F=', {'foo/': ''}),
            ('+', {' ': True}),
            ('%20=', {' ': ''}),
            ('foo&foo&foo=&foo=bar', {'foo': [True, True, '', 'bar']}),
            ('foo&bar=bar', {'foo': True, 'bar': 'bar'}),
        )
        for qs, expected in test_cases:
            actual = self.signer._query_string_to_args(qs)
            self.assertEqual(expected, actual)

    def test_args_to_query_string(self):
        test_cases = (
            ({}, ''),
            ({'foo/': True}, 'foo%2F'),
            ({' ': ' '}, '%20=%20'),
            ({'?': ['/', ' ', True]}, '%3F=%2F&%3F=%20&%3F'),
            ({'foo': True, 'bar': ''}, 'foo&bar='),
            ({'foo': [True, '', True], 'bar': True}, 'bar&foo&foo=&foo'),
        )
        for args, expected in test_cases:
            actual = self.signer._args_to_query_string(args)
            self.assertEqual(expected, actual)

    def test_get_request_date(self):
        test_cases = (
            (1481026060, '20161206T120740Z'),
            (1481026060.00, '20161206T120740Z'),
            ('20161206T120741Z', '20161206T120741Z'),
        )

        for request_date_in, expected in test_cases:
            actual = self.signer._get_request_date(request_date_in)
            self.assertEqual(expected, actual)

    def test_clean_query_string(self):
        test_cases = (
            ('', ''),
            ('foo', 'foo'),
            ('X-Amz-Date', ''),
            ('X-Amz-Date=', ''),
            ('X-Amz-Date=bar', ''),
            ('foo&X-Amz-Date=bar&bar', 'foo&bar'),
            ('foo=&' +
             'X-Amz-Algorithm=AWS4-HMAC-SHA256&' +
             'X-Amz-Credential=m%2Faws4_request&' +
             'X-Amz-Date=20150830T123600Z&' +
             'X-Amz-Expires=60&' +
             'X-Amz-SignedHeaders=content-type%3Bhost&' +
             'X-Amz-Signature=81d02', 'foo='),
        )
        for qs, expected in test_cases:
            actual = self.signer._clean_query_string(qs)
            self.assertEqual(expected, actual)

    def test_clean_args(self):
        test_cases = (
            ({}, {}),
            ({'foo': 'bar'}, {'foo': 'bar'}),
            ({'foo': ['bar', True]}, {'foo': ['bar', True]}),
            ({'foo': ['bar', True], 'X-Amz-Date': ''}, {'foo': ['bar', True]}),
            ({'foo': ['bar', True], 'X-Amz-Date': True},
             {'foo': ['bar', True]}),
            ({'foo': ['bar', True], 'X-Amz-Date': [True]},
             {'foo': ['bar', True]}),
            ({'foo': ['bar', True],
              'X-Amz-Date': [True],
              'X-Amz-Algorithm': [''],
              'X-Amz-Credential': 'foo',
              'X-Amz-Expires': 80,
              'X-Amz-SignedHeaders': 'foo;bar',
              'X-Amz-Signature': ' '}, {'foo': ['bar', True]}),
        )
        for args, expected in test_cases:
            actual = self.signer._clean_args(args)
            self.assertEqual(expected, actual)

    def test_trimall(self):
        test_cases = (
            ('', ''),
            ('foo', 'foo'),
            ('  foo', 'foo'),
            ('  foo   ', 'foo'),
            ('  f oo   ', 'f oo'),
            ('  f o   o   ', 'f o o'),
            ('  ~!@#$%^&*()_+   ', '~!@#$%^&*()_+'),
            ('  f , oo<>?,./   ', 'f , oo<>?,./'),
            ('  :    ;  ', ': ;'),
        )
        for str, expected in test_cases:
            actual = self.signer._trimall(str)
            self.assertEqual(expected, actual)

    def test_standardize_headers(self):
        test_cases = (
            ({}, [], '', {}),
            ({'foo': 'bar'}, [], 'foo', {'foo': 'bar'}),
            ({'foo': 'bar'}, ['Foo'], '', {'foo': 'bar'}),
            ({'Foo': 'BAR'}, ['foo'], '', {'foo': 'BAR'}),
            ({'Foo  ': 'BAR'}, ['foo'], '', {'foo': 'BAR'}),
            ({'F  oo  ': 'BAR'}, ['F  oo'], '', {'f  oo': 'BAR'}),
            ({'F  oo  ': '  BA  R  '}, [], 'f  oo', {'f  oo': 'BA R'}),
            ({'foo': 'bar', 'bar': '   '}, [],
             'bar;foo', {'foo': 'bar', 'bar': ''}),
        )
        for headers, not_to_sign_headers, expected_signed_headers, expected_stand_headers in test_cases:
            actual_signed_headers, actual_stand_headers = self.signer._standardize_headers(
                headers, not_to_sign_headers)
            self.assertEqual(expected_signed_headers, actual_signed_headers)
            self.assertEqual(expected_stand_headers, actual_stand_headers)

    def test_modify_request_headers(self):
        test_cases = (
            (1,
             {'headers': {}},
             False,
             False,
             '20161206T000000Z',
             'UNSIGNED-PAYLOAD',
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': 'UNSIGNED-PAYLOAD'}),

            (2,
             {'headers': {'foo': 'bar'}},
             False,
             True,
             '20161206T000000Z',
             EMPTY_PAYLOAD_HASH,
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': EMPTY_PAYLOAD_HASH,
              'foo': 'bar'}),

            (3,
             {'headers': {},
              'body': 'foo'},
             False,
             True,
             '20161206T000000Z',
             '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae'}),

            (4,
             {'headers': {'X-Amz-Date': '000000T000000Z'}},
             False,
             True,
             '20161206T000000Z',
             EMPTY_PAYLOAD_HASH,
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': EMPTY_PAYLOAD_HASH}),

            (5,
             {'headers': {'Date': 'Wed, 07 Dec 2016 05:20:11 GMT'}},
             False,
             True,
             '20161206T000000Z',
             EMPTY_PAYLOAD_HASH,
             {'X-Amz-Date': '20161206T000000Z',
              'Date': 'Wed, 07 Dec 2016 05:20:11 GMT',
              'X-Amz-Content-SHA256': EMPTY_PAYLOAD_HASH}),

            (6,
             {'headers': {'X-Amz-Content-SHA256': 'foo'}},
             False,
             True,
             '20161206T000000Z',
             'foo',
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': 'foo'}),

            (7,
             {'headers': {'X-Amz-Content-SHA256': 'foo'},
              'body': 'bar'},
             False,
             True,
             '20161206T000000Z',
             'foo',
             {'X-Amz-Date': '20161206T000000Z',
              'X-Amz-Content-SHA256': 'foo'}),

            (8,
             {'headers': {}},
             True,
             True,
             '20161206T000000Z',
             'UNSIGNED-PAYLOAD',
             {}),

            (9,
             {'headers': {'X-Amz-Date': 'foo'}},
             True,
             True,
             '20161206T000000Z',
             'UNSIGNED-PAYLOAD',
             {'X-Amz-Date': '20161206T000000Z'}),

            (10,
             {'headers': {'Date': 'foo'}},
             True,
             True,
             '20161206T000000Z',
             'UNSIGNED-PAYLOAD',
             {'Date': 'foo'}),

            (11,
             {'headers': {'X-Amz-Content-SHA256': 'foo'},
              'body': 'bar'},
             True,
             True,
             '20161206T000000Z',
             'foo',
             {'X-Amz-Content-SHA256': 'foo'}),
        )

        for i, request, presign, sign_payload, request_date, expected_hashed_payload, expected_headers in test_cases:
            actual_hashed_payload = self.signer._modify_request_headers(
                request, presign, sign_payload, request_date)
            self.assertEqual(expected_hashed_payload, actual_hashed_payload,
                             'in case %d: %s, %s' % (i, expected_hashed_payload,
                                                     actual_hashed_payload))
            self.assertEqual(expected_headers, request['headers'],
                             'in case %d: %s, %s' % (i, repr(expected_headers),
                                                     repr(request['headers'])))

    def test_use_both_args_and_query_string(self):
        test_cases = (
            {'verb': 'GET',
             'uri': '/?',
             'args': {},
             'headers': {}, },

            {'verb': 'GET',
             'uri': '/foo/bar?foo',
             'args': {'foo': 'bar'},
             'headers': {}, },

            {'verb': 'GET',
             'uri': '/?',
             'args': {'foo': 'bar'},
             'headers': {}, },
        )
        for request in test_cases:
            with self.assertRaises(Exception):
                self.signer.add_auth(request)

    def test_custom_signing_date(self):
        request = {
            'verb': 'GET',
            'uri': '/',
            'headers': {
                'Host': '127.0.0.1',
            },
        }

        ctx = self.signer.add_auth(request, signing_date='20150102')
        self.assertEqual('20150102/us-east-1/s3/aws4_request',
                         ctx['credential_scope'])



import logging
import os
import time

from pykit import fsutil
from pykit.cgrouparch import cgroup_util
from pykit.cgrouparch import model
from pykit.cgrouparch import util

logger = logging.getLogger(__name__)


def get_pids_of_one_pid_file(pid_file):
    try:
        pid = util.get_pid_from_file(pid_file)
        pids = util.get_all_pids(pid)
        logger.info('pids for pid file: %s is %s' % (pid_file, pids))
        return pids

    except Exception as e:
        logger.exception('failed to get pids of pid file: %s, %s' %
                         (pid_file, repr(e)))
        return []


def get_cgroup_pids(pid_files):
    cgroup_pids = []

    for pid_file in pid_files:
        pids = get_pids_of_one_pid_file(pid_file)
        cgroup_pids.extend(pids)

    return cgroup_pids


def update_one_cgroup_pids(cgroup_path, cgroup_conf, context):
    if 'conf' in cgroup_conf:
        cgroup_name = os.path.split(cgroup_path)[-1]
        try:
            pid_files = context['get_cgroup_pid_file'](cgroup_name)
            if pid_files is not None:
                cgroup_conf['conf']['pids'] = get_cgroup_pids(pid_files)

        except Exception as e:
            logger.exception('failed to get pid files of: %s, %s' %
                             (cgroup_name, repr(e)))

    sub_cgroup = cgroup_conf.get('sub_cgroup')
    if sub_cgroup is None:
        return

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_cgroup_path = os.path.join(cgroup_path, sub_cgroup_name)

        update_one_cgroup_pids(sub_cgroup_path, sub_cgroup_conf, context)

    return


def update_cgroup_pids(context):
    arch_conf = context['arch_conf']['value']

    for subsystem_name, subsystem_arch_conf in arch_conf.iteritems():
        cgroup_path = os.path.join(context['cgroup_dir'], subsystem_name)
        cgroup_conf = subsystem_arch_conf

        try:
            update_one_cgroup_pids(cgroup_path, cgroup_conf, context)
        except Exception as e:
            logger.exception('failed to update pids of cgroup: %s, %s' %
                             (cgroup_path, repr(e)))
    return


def build_cgroup_arch(subsystem_dir, cgroup_path, cgroup_conf,
                      protected_cgroup):
    sub_dirs = fsutil.get_sub_dirs(cgroup_path)

    sub_cgroup = cgroup_conf.get('sub_cgroup', {})

    for sub_dir in sub_dirs:
        if sub_dir not in sub_cgroup:
            path = os.path.join(cgroup_path, sub_dir)
            logger.warn('unknown cgroup: %s' % path)

            if sub_dir in protected_cgroup:
                continue

            cgroup_util.remove_cgroup(subsystem_dir, path)
            logger.info('remove unknown cgroup: %s', path)

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_cgroup_path = os.path.join(cgroup_path, sub_cgroup_name)

        if sub_cgroup_name not in sub_dirs:
            cgroup_util.create_cgroup(sub_cgroup_path)
            logger.info('create cgroup: %s', sub_cgroup_path)

        build_cgroup_arch(subsystem_dir, sub_cgroup_path,
                          sub_cgroup_conf, protected_cgroup)

    return


def build_all_subsystem_cgroup_arch(context):
    arch_conf = context['arch_conf']['value']

    for subsystem_name, subsystem_arch_conf in arch_conf.iteritems():
        subsystem_dir = os.path.join(context['cgroup_dir'], subsystem_name)

        cgroup_path = subsystem_dir
        cgroup_conf = subsystem_arch_conf

        protected_cgroup = context.get('protected_cgroup')
        if protected_cgroup is None:
            protected_cgroup = []

        build_cgroup_arch(subsystem_dir, cgroup_path, cgroup_conf,
                          protected_cgroup)


def set_one_cgroup(subsystem_model, cgroup_path, cgroup_conf):
    if 'conf' in cgroup_conf:
        subsystem_model['set_cgroup'](cgroup_path, cgroup_conf['conf'])

    sub_cgroup = cgroup_conf.get('sub_cgroup')
    if sub_cgroup is None:
        return

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_cgroup_path = os.path.join(cgroup_path, sub_cgroup_name)

        set_one_cgroup(subsystem_model, sub_cgroup_path, sub_cgroup_conf)

    return


def set_cgroup(context):
    update_cgroup_pids(context)

    arch_conf = context['arch_conf']['value']

    for subsystem_name, subsystem_arch_conf in arch_conf.iteritems():
        cgroup_path = os.path.join(context['cgroup_dir'], subsystem_name)
        cgroup_conf = subsystem_arch_conf

        subsystem_model = model.subsystem[subsystem_name]

        try:
            set_one_cgroup(subsystem_model, cgroup_path, cgroup_conf)
        except Exception as e:
            logger.exception('failed to set cgroup: %s, %s' %
                             (cgroup_path, repr(e)))
    return


def loop_set_cgroup(context):
    while True:
        start_time = time.time()

        set_cgroup(context)

        time_used = time.time() - start_time
        logger.info('set cgroup at: %f, time used: %f' %
                    (start_time, time_used))

        time.sleep(context['tasks_update_interval'])


def reset_statistics_one_cgroup(subsystem_model, cgroup_path, cgroup_conf):
    subsystem_model['reset_statistics'](cgroup_path)
    logger.info('reset statistics of cgroup: %s' % cgroup_path)

    sub_cgroup = cgroup_conf.get('sub_cgroup')
    if sub_cgroup is None:
        return

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_cgroup_path = os.path.join(cgroup_path, sub_cgroup_name)

        reset_statistics_one_cgroup(subsystem_model, sub_cgroup_path,
                                    sub_cgroup_conf)
    return


def reset_statistics(context):
    arch_conf = context['arch_conf']['value']

    for subsystem_name, subsystem_arch_conf in arch_conf.iteritems():
        cgroup_path = os.path.join(context['cgroup_dir'], subsystem_name)
        cgroup_conf = subsystem_arch_conf

        subsystem_model = model.subsystem[subsystem_name]

        try:
            reset_statistics_one_cgroup(subsystem_model, cgroup_path,
                                        cgroup_conf)
        except Exception as e:
            logger.exception('failed to reset statistics of cgroup: %s, %s' %
                             (cgroup_path, repr(e)))
    return



#!/usr/bin/env python2
# coding: utf-8

import gc
import os
import socket
import ssl
import threading
import time
import unittest
from BaseHTTPServer import BaseHTTPRequestHandler
from BaseHTTPServer import HTTPServer

from pykit import http
from pykit import ututil

dd = ututil.dd


HOST = '127.0.0.1'
PORT = 38002
KB = 1024
MB = (1024**2)
HOME_PATH = os.path.dirname(os.path.abspath(__file__))


class TestHttpClient(unittest.TestCase):

    special_cases = {
        'test_recving_server_close':
        (0, 1, 'HTTP/1.1 200 OK\r\nContent-Length: 10\r\n\r\n'),

        'test_server_delay_response':
        (0.5, 1, 'HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nabcd'),

        'test_raise_chunked_size_error':
        (0, 10, 'HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\nfoo\r\n'),

        'test_raise_socket_timeout':
        (3, 1, 'H'),

        'test_raise_line_too_long_error':
        (0, KB, 'a' * 65536),

        'test_request_chunked':
        (),

        'test_readlines':
        (0, 10, 'HTTP/1.1 200 OK\r\nContent-Length: 131086\r\n\r\n' + 'a'*65540 + '\r\nbb\r\n' + 'c'*65540),

        'test_readlines_delimiter':
        (0, 10, 'HTTP/1.1 200 OK\r\nContent-Length: 15\r\n\r\nabcd\rbcde\rcdef\r'),

    }
    request_headers = {}
    request_body = {}

    def test_raise_connnect_error(self):

        h = http.Client(HOST, PORT)
        self.assertRaises(http.NotConnectedError, h.send_body, None)

    def test_raise_line_too_long_error(self):

        h = http.Client(HOST, PORT)
        self.assertRaises(http.LineTooLongError,
                          h.request, '/line_too_long')

    def test_raise_response_headers_error(self):

        cases = (
            '/invalid_content_len',
            '/invalid_header',
        )
        h = http.Client(HOST, PORT)
        for uri in cases:
            self.assertRaises(http.HeadersError, h.request, uri)

    def test_raise_chunked_size_error(self):

        h = http.Client(HOST, PORT)
        h.request('')
        self.assertRaises(http.ChunkedSizeError, h.read_body, 10)

    def test_raise_response_not_ready_error(self):

        h = http.Client(HOST, PORT)
        self.assertRaises(http.ResponseNotReadyError, h.read_response)

    def test_raise_socket_timeout(self):

        h = http.Client(HOST, PORT, 2)
        self.assertRaises(socket.timeout, h.request, '')

    def test_raise_badstatus_line_error(self):

        cases = (
            '/invalid_line',
            '/invalid_protocol',
            '/<100',
            '/>999',
        )

        h = http.Client(HOST, PORT)
        for uri in cases:

            self.assertRaises(http.BadStatusLineError, h.request, uri)

    def test_raise_socket_error(self):

        h = http.Client(HOST, PORT)
        h.request('/socket_error')
        self.assertRaises(socket.error, h.read_body, 10)

    def test_get_http_request(self):

        cases = (
            ('/get_1b', 1, 'a', (), False),
            ('/get_1b', 10, 'a', (), False),
            ('/get_1b', None, 'a', (), False),
            ('/get_10k', KB, 'bc' * 5 * KB, (), False),
            ('/get_10k', 20 * KB, 'bc' * 5 * KB, (), False),
            ('/get_10k', None, 'bc' * 5 * KB, (), False),
            ('/get_30m', 10 * MB, 'cde' * 10 * MB, (), False),
            ('/get_30m', 50 * MB, 'cde' * 10 * MB, (), False),
            ('/get_30m', None, 'cde' * 10 * MB, (), False),

            ('/get_10b_chunked', 1, 'f' * 10, (), True),
            ('/get_10b_chunked', 10, 'f' * 10, (), True),
            ('/get_10b_chunked', None, 'f' * 10, (), True),
            ('/get_10k_chunked', KB, 'gh' * 5 * KB, (), True),
            ('/get_10k_chunked', 20 * KB, 'gh' * 5 * KB, (), True),
            ('/get_10k_chunked', None, 'gh' * 5 * KB, (), True),
            ('/get_30m_chunked', 10 * MB, 'ijk' * 10 * MB, (), True),
            ('/get_30m_chunked', 50 * MB, 'ijk' * 10 * MB, (), True),
            ('/get_30m_chunked', None, 'ijk' * 10 * MB, (), True),

            ('/get_10b_range', 1, 'l' * 10, (2, 8), False),
            ('/get_10b_range', 10, 'l' * 10, (2, 8), False),
            ('/get_10b_range', None, 'l' * 10, (2, 8), False),
            ('/get_10k_range', KB, 'mn' * 5 * KB, (KB, 8 * KB), False),
            ('/get_10k_range', 20 * KB, 'mn' * 5 * KB, (KB, 8 * KB), False),
            ('/get_10k_range', None, 'mn' * 5 * KB, (KB, 8 * KB), False),
            ('/get_30m_range', 10 * MB, 'opq' * 10 * MB, (2 * MB, 25 * MB), False),
            ('/get_30m_range', 50 * MB, 'opq' * 10 * MB, (2 * MB, 25 * MB), False),
            ('/get_30m_range', None, 'opq' * 10 * MB, (2 * MB, 25 * MB), False),
        )
        h = http.Client(HOST, PORT)
        for uri, each_read_size, expected_res, content_range, chunked in cases:
            h.request(uri)

            bufs = ''
            if each_read_size is None:
                bufs = h.read_body(None)
                self.assertEqual(h.has_read, len(bufs))
            else:
                while True:
                    b = h.read_body(each_read_size)
                    if len(b) <= 0:
                        break
                    bufs += b
                    self.assertEqual(h.has_read, len(bufs))

            start, end = 0, len(expected_res)
            if len(content_range) >= 2:
                start, end = content_range[0], content_range[1] + 1

            self.assertEqual(expected_res[start:end], bufs)
            self.assertEqual(None if chunked else len(bufs), h.content_length)
            self.assertEqual(chunked, h.chunked)

    def test_status(self):

        cases = (
            ('/get_200', 200),
            ('/get_304', 304),
            ('/get_404', 404),
            ('/get_500', 500),
        )

        h = http.Client(HOST, PORT)
        for uri, expected_status in cases:
            h.request(uri)

            self.assertEqual(expected_status, h.status)

    def test_request_chunked(self):

        h = http.Client(HOST, PORT)
        h.send_request('', 'PUT', {'Transfer-Encoding': 'chunked'})

        cases = (
            ('aaaaaaaaa', 100),
            ('bbbbbbbbbbbbbb', 100),
            ('0000000000000', 100),
            ('200_status', 200)
                )

        for body, status in cases:
            h.send_body(body)
            self.assertEqual(h.read_status(False), status)

    def test_request_headers(self):

        cases = (
            ('/header_1', {'host': 'example.com'}),
            ('/header_2', {'host': 'example.com', 'b': 'bar'}),
            ('/header_3', {'host': 'example.com', 'b': 'bar', 'f': 'foo'}),
        )

        h = http.Client(HOST, PORT)
        for uri, headers in cases:
            h.request(uri, headers=headers)
            time.sleep(0.1)

            self.assertEqual(headers, self.request_headers)

    def test_response_headers(self):

        cases = (
            ('/header_res1', {'f': 'foo'}),
            ('/header_res2', {'f': 'foo', 'b': 'bar'}),
            ('/header_res3', {'f': 'foo', 'b': 'bar', 't': 'too'}),
        )

        h = http.Client(HOST, PORT)
        for uri, expected_headers in cases:
            h.request(uri)

            self.assertEqual(expected_headers, h.headers)

    def test_send_body(self):

        cases = (
            ('/put_1b', 'a', {'Content-Length': 1}),
            ('/put_10k', 'bc' * 5 * KB, {'Content-Length': 10 * KB}),
            ('/put_30m', 'cde' * 10 * MB, {'Content-Length': 30 * MB}),
        )

        h = http.Client(HOST, PORT)
        for uri, body, headers in cases:
            h.send_request(uri, method='PUT', headers=headers)
            h.send_body(body)
            h.read_response()
            time.sleep(0.1)

            self.assertEqual(body, self.request_body)

    def test_readlines(self):

        h = http.Client(HOST, PORT)
        h.request('')

        expected_body = ('a' * 65540 + '\r\n', 'bb\r\n', 'c' * 65540)
        for idx, line in enumerate(h.readlines()):
            self.assertEqual(expected_body[idx], line)

    def test_readlines_delimiter(self):

        h = http.Client(HOST, PORT)
        h.request('')

        expected_body = ('abcd\r', 'bcde\r', 'cdef\r')
        for idx, line in enumerate(h.readlines('\r')):
            self.assertEqual(expected_body[idx], line)

    def test_recving_server_close(self):

        h = http.Client(HOST, PORT, 3)
        succ = False

        try:
            h.request('')
            h.read_body(1024)
        except socket.error as e:
            dd(repr(e) + ' while recv server close')
            succ = True
        except Exception as e:
            dd(repr(e) + ' unexpected exception')

        self.assertTrue(succ)

    def test_server_delay_response(self):

        case = ({'content-length': '4'}, 'abcd')
        expected_headers, expected_body = case

        h = http.Client(HOST, PORT, 1)
        h.request('')
        body = h.read_body(1024)

        self.assertEqual(expected_headers, h.headers)
        self.assertEqual(expected_body, body)

    def test_client_delay_send_data(self):

        case = ('/client_delay', {'Content-Length': 10}, 'abcde' * 2)
        uri, headers, body = case

        h = http.Client(HOST, PORT, 3)
        h.send_request(uri, method='PUT', headers=headers)

        while len(body) > 0:
            h.send_body(body[:1])
            time.sleep(1)
            body = body[1:]

        self.assertEqual(case[2], self.request_body)

    def test_garbage_collector(self):

        h = http.Client(HOST, PORT)
        h.request('/get_30m')
        h.read_body(None)
        del h

        gc.collect()
        self.assertListEqual([], gc.garbage)

    def test_trace(self):

        class FakeErrorDuringHTTP(Exception):
            pass

        h = http.Client(HOST, PORT)
        h.request('/get_10k')
        h.read_body(1)
        h.read_body(None)

        # emulate error
        try:
            with h.stopwatch.timer('exception'):
                raise FakeErrorDuringHTTP(3)
        except Exception:
            pass

        trace = h.get_trace()
        dd('trace:', trace)

        ks = (
            'conn',
            'send_header',
            'recv_status',
            'recv_header',
            'recv_body',
        )

        for i, k in enumerate(ks):
            self.assertEqual(k, trace[i]['name'])
            self.assertEqual(type(0.1), type(trace[i]['time']))

        names = [x['name'] for x in trace]
        self.assertEqual(['conn',
                          'send_header',
                          'recv_status',
                          'recv_header',
                          'recv_body',
                          'recv_body',
                          'exception',
                          'pykit.http.Client'],
                         names)

        dd('trace str:', h.get_trace_str())

    def test_trace_min_tracing_milliseconds(self):

        h = http.Client(HOST, PORT, stopwatch_kwargs={
                        'min_tracing_milliseconds': 1000})
        h.request('/get_10k')
        h.read_body(None)

        # only steps cost time>1000 are traced. thus nothing should be traced
        trace_str = h.get_trace_str()
        dd('trace:', trace_str)

        self.assertEqual('', trace_str)

        self.assertEqual([], h.get_trace())

    def test_https(self):
        cases = (
            ('/get_1b', 'a'),
            ('/get_10k', 'bc' * 5 * KB),
            ('/get_30m', 'cde' * 10 * MB),
        )

        context = ssl._create_unverified_context()
        cli = http.Client(HOST, PORT, https_context=context)
        for uri, expected_res in cases:
            cli.request(uri)
            body = cli.read_body(None)

            self.assertEqual(200, cli.status)
            self.assertEqual(expected_res, body)

    def __init__(self, *args, **kwargs):

        super(TestHttpClient, self).__init__(*args, **kwargs)
        self.server_thread = None
        self.http_server = None

    def setUp(self):

        self.server_thread = threading.Thread(target=self._start_server)
        self.server_thread.daemon = True
        self.server_thread.start()
        time.sleep(0.1)

    def tearDown(self):

        if self.http_server is not None:
            self.http_server.shutdown()
            self.http_server.server_close()

        self.server_thread.join()

    def _start_server(self):

        if self._testMethodName in self.special_cases:
            self._special_case_handle()
        else:
            addr = (HOST, PORT)
            self.http_server = HTTPServer(addr, Handle)
            if 'https' in self._testMethodName:
                cert_file = os.path.join(HOME_PATH, 'test_https.pem')
                self.http_server.socket = ssl.wrap_socket(self.http_server.socket,
                                                          certfile=cert_file,
                                                          server_side=True)
            self.http_server.serve_forever()

    def _special_case_handle(self):

        addr = (HOST, PORT)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(addr)
        sock.listen(10)


        if self._testMethodName == 'test_request_chunked':

            conn, _ = sock.accept()
            for i in range(3):
                data = conn.recv(1024)
                dd('recv data:' + data)

                res = 'HTTP/1.1 100 CONTINUE\r\n\r\n'
                conn.sendall(res)

            data = conn.recv(1024)
            dd('recv data:' + data)
            res = 'HTTP/1.1 200 OK\r\n\r\n'
            conn.sendall(res)

        else:
            conn, _ = sock.accept()
            data = conn.recv(1024)
            dd('recv data:' + data)
            res = self.special_cases.get(self._testMethodName)
            if res is None:
                return

            sleep_time, each_send_size, content = res
            try:
                while len(content) > 0:
                    conn.sendall(content[:each_send_size])
                    content = content[each_send_size:]
                    time.sleep(sleep_time)
            except socket.error as e:
                dd(repr(e) + ' while response')

        time.sleep(1)
        conn.close()
        sock.close()


class Handle(BaseHTTPRequestHandler):

    all_responses = {
        '/invalid_content_len': (200, {'content-length': 'abc'}, (0, '')),
        '/invalid_header': (200, {}, (0, '')),

        '/get_1b': (200, {'content-length': 1}, (1, 'a')),
        '/get_10k': (200, {'content-length': 10 * KB}, (10240, 'bc' * 5 * KB)),
        '/get_30m': (200, {'content-length': 30 * MB}, (10 * MB, 'cde' * 10 * MB)),

        '/get_10b_chunked': (200, {'Transfer-Encoding': 'chunked'}, (5, 'f' * 10)),
        '/get_10k_chunked': (200, {'Transfer-Encoding': 'chunked'}, (KB, 'gh' * 5 * KB)),
        '/get_30m_chunked': (200, {'Transfer-Encoding': 'chunked'}, (10 * MB, 'ijk' * 10 * MB)),

        '/get_10b_range': (206,
                           {'Content-Range': 'bytes %d-%d/%d' % (2, 8, 10),
                            'Content-Length': 7},
                           (5, 'l' * 10)),
        '/get_10k_range': (206,
                           {'Content-Range': 'bytes %d-%d/%d' % (KB, 8 * KB, 10 * KB),
                            'Content-Length': 7 * KB + 1},
                           (KB, 'mn' * 5 * KB)),
        '/get_30m_range': (206,
                           {'Content-Range': 'bytes %d-%d/%d' % (2 * MB, 25 * MB, 30 * MB),
                            'Content-Length': 23 * MB + 1},
                           (10 * MB, 'opq' * 10 * MB)),

        '/get_200': (200, {}, (0, '')),
        '/get_304': (304, {}, (0, '')),
        '/get_404': (404, {}, (0, '')),
        '/get_500': (500, {}, (0, '')),

        '/header_1': (200, {}, (0, '')),
        '/header_2': (200, {}, (0, '')),
        '/header_3': (200, {}, (0, '')),

        '/header_res1': (200, {'f': 'foo'}, (0, '')),
        '/header_res2': (200, {'f': 'foo', 'b': 'bar'}, (0, '')),
        '/header_res3': (200, {'f': 'foo', 'b': 'bar', 't': 'too'}, (0, '')),

        '/invalid_line': (200, {}, (0, '')),
        '/invalid_protocol': (200, {}, (0, '')),
        '/<100': (10, {}, (0, '')),
        '/>999': (1000, {}, (0, '')),

        '/socket_error': (200, {'Content-Length': 10}, (0, '')),
    }

    def send_response(self, code, message=None):

        self.log_request(code)
        if message is None:
            if code in self.responses:
                message = self.responses[code][0]
            else:
                message = ''
        if self.request_version != 'HTTP/0.9':
            if self.path == '/invalid_protocol':
                protocol = 'foo'
            elif self.path == '/invalid_line':
                self.wfile.write(self.protocol_version + '\r\n')
                return
            else:
                protocol = self.protocol_version
            self.wfile.write("%s %d %s\r\n" %
                             (protocol, code, message))
            if self.path == '/invalid_header':
                self.wfile.write('foo\r\n')

    def do_PUT(self):

        try:
            length = int(self.headers.getheader('Content-Length'))
        except (TypeError, ValueError) as e:
            dd(repr(e))
            return

        read_bytes = 0
        bufs = ''

        try:
            while read_bytes < length:
                bufs += self.rfile.read(length - read_bytes)
                read_bytes = len(bufs)

            TestHttpClient.request_body = bufs
            self.send_response(200)
            self.send_header('Content-Length', 0)
            self.end_headers()
        except Exception as e:
            dd(repr(e) + ' while parse put request')

    def do_GET(self):

        TestHttpClient.request_headers = self.headers.dict

        res = self.all_responses.get(self.path)
        if res is None:
            dd('path error:' + self.path)
            return

        status, headers, body = res

        try:
            self.send_response(status)

            for k, v in headers.items():
                self.send_header(k, v)
            self.end_headers()

            self._send_body(headers, body)
        except Exception as e:
            dd(repr(e) + ' while parse get request')

    def _send_body(self, headers, body):

        each_send_size, data = self._get_body(headers, body)
        ext = ';extname'
        while len(data) > 0:
            send_buf = data[:each_send_size]
            if 'Transfer-Encoding' in headers:
                if len(ext) > 0:
                    ext = ''
                else:
                    ext = ';extname'
                send_buf = '%x%s\r\n%s\r\n' % (len(send_buf), ext, send_buf)

            self.wfile.write(send_buf)
            data = data[each_send_size:]

        if 'Transfer-Encoding' in headers:
            self.wfile.write('0\r\n\r\n')

    def _get_body(self, headers, body):

        each_send_size, data = body
        start = 0
        end = len(data)
        if 'Content-Range' in headers:
            val = headers['Content-Range'][6:]
            val = val[:val.find('/')]
            start, end = val.split('-', 1)

        return each_send_size, data[int(start):int(end) + 1]



#!/usr/bin/env python2
# coding: utf-8

import logging
from collections import OrderedDict

from geventwebsocket import Resource
from geventwebsocket import WebSocketApplication
from geventwebsocket import WebSocketServer

from pykit import utfjson
from pykit.cgrouparch import account

global_value = {}

logger = logging.getLogger(__name__)


class CgroupArchWebSocketApplication(WebSocketApplication):

    def on_open(self):
        logger.info('on open')

    def on_message(self, message_str):
        if message_str is None:
            return

        try:
            self.process_message(message_str)
        except Exception as(e):
            logger.exception('failed to process message: ' + repr(e))
            self.send_json({'error': repr(e)})

    def on_close(self, reason):
        logger.info('on close')

    def process_message(self, message_str):
        message = utfjson.load(message_str)

        cmd = message['cmd']
        args = message.get('args')
        if args is None:
            args = {}

        result = self.do_cmd(cmd, args)
        self.send_json(result)

    def do_cmd(self, cmd, args):
        if cmd == 'show_account':
            return self.show_account(args)
        elif cmd == 'get_conf':
            return global_value['context']['arch_conf']
        else:
            return {'error': 'invalid cmd: %s' % cmd}

    def show_account(self, args):
        return account.show(global_value['context'], args)

    def send_json(self, value):
        value_str = utfjson.dump(value)
        self.ws.send(value_str)


def run(context, ip='0.0.0.0', port=22348):
    global_value['context'] = context
    WebSocketServer(
        (ip, port),
        Resource(OrderedDict({'/': CgroupArchWebSocketApplication})),
    ).serve_forever()



#!/usr/bin/env python2
# coding: utf-8

import os
import time
import unittest

from pykit import fsutil
from pykit import humannum
from pykit import proc
from pykit import threadutil
from pykit import ututil

dd = ututil.dd

# xx/pykit/fsutil/test/
this_base = os.path.dirname(__file__)

pyt = 'python2'


class TestFSUtil(unittest.TestCase):

    def test_exceptions(self):

        dd('present: ', fsutil.FSUtilError)
        dd('present: ', fsutil.NotMountPoint)

    def test_get_all_mountpoint(self):
        mps = fsutil.get_all_mountpoint()
        dd('mount points:', mps)
        self.assertIn('/', mps)

        mpsall = fsutil.get_all_mountpoint(all=True)
        dd('all mount points:', mps)

        self.assertTrue(len(mpsall) > len(mps))
        self.assertEqual(set([]), set(mps) - set(mpsall))

    def test_mountpoint(self):

        cases = (
                ('/',               '/',),
                ('/bin/ls',         '/',),
                ('/dev',            '/dev',),
                ('/dev/',           '/dev',),
                ('/dev/random',     '/dev',),
                ('/dev/inexistent', '/dev',),
        )

        for path, expected in cases:

            dd(path, ' --> ', expected)

            rst = fsutil.get_mountpoint(path)
            dd('rst: ', rst)

            self.assertEqual(expected, rst)

            # test assert_mount_point

            fsutil.assert_mountpoint(rst)
            self.assertRaises(fsutil.NotMountPoint,
                              fsutil.assert_mountpoint, path + '/aaa')

    def test_get_disk_partitions(self):

        rst = fsutil.get_disk_partitions()

        self.assertIn('/', rst)

        root = rst['/']
        for k in ('device', 'mountpoint', 'fstype', 'opts'):
            self.assertIn(k, root)

        notall = fsutil.get_disk_partitions(all=False)
        self.assertTrue(len(rst) > len(notall))
        self.assertEqual(set([]), set(notall) - set(rst))

    def test_get_device(self):

        rst = fsutil.get_device('/inexistent')

        # GNU "grep -o" split items into several lines
        root_dev = proc.shell_script(
            'mount | grep " on / " | grep -o "[^ ]*" | head -n1')[1].strip()

        self.assertEqual(root_dev, rst)

    def test_get_device_fs(self):

        # GNU "grep -o" split items into several lines
        rc, out, err = proc.shell_script(
            'mount | grep " on / " | grep -o "[^ ]*" | head -n1')
        dd('find device on /', rc, out, err)

        dev = out.strip()
        dd('device: ', dev)
        rst = fsutil.get_device_fs(dev)
        self.assertIn(rst, ('hfs', 'xfs', 'ext2', 'ext3', 'ext4'))

    def test_get_path_fs(self):

        rst = fsutil.get_path_fs('/dev')
        dd('fs of /dev: ', rst)
        self.assertNotEqual('unknown', rst)
        self.assertTrue('dev' in rst)

        rst = fsutil.get_path_fs('/blabla')
        dd('fs of /blabla: ', rst)
        self.assertIn(rst, ('hfs', 'xfs', 'ext2', 'ext3', 'ext4'))

    def test_get_path_usage(self):

        path = '/'
        rst = fsutil.get_path_usage(path)

        # check against os.statvfs.
        st = os.statvfs(path)

        dd(humannum.humannum(rst))

        # space is changing..

        self.assertAlmostEqual(st.f_frsize * st.f_bavail, rst['available'], delta=4 * 1024**2)
        self.assertAlmostEqual(st.f_frsize * st.f_blocks, rst['total'], delta=4 * 1024**2)
        self.assertAlmostEqual(st.f_frsize * (st.f_blocks - st.f_bavail), rst['used'], delta=4 * 1024**2)
        self.assertAlmostEqual((st.f_blocks - st.f_bavail) * 100 / st.f_blocks,
                               int(rst['percent'] * 100), delta=4 * 1024**2)

        if st.f_bfree > st.f_bavail:
            self.assertLess(rst['available'], st.f_frsize * st.f_bfree)
        else:
            dd('st.f_bfree == st.f_bavail')

        # check against df

        rc, out, err = proc.shell_script('df -m / | tail -n1')
        # Filesystem 1M-blocks   Used Available Capacity  iused    ifree %iused  Mounted on
        # /dev/disk1    475828 328021    147556    69% 84037441 37774557   69%   /
        dd('space of "/" from df')
        total_mb = out.strip().split()[1]

        # "123.8M", keep int only
        rst_total_mb = humannum.humannum(rst['total'], unit=humannum.M)
        rst_total_mb = rst_total_mb.split('.')[0].split('M')[0]

        dd('total MB from df:', total_mb, 'result total MB:', rst_total_mb)

        self.assertAlmostEqual(int(total_mb), int(rst_total_mb), delta=4)

    def test_get_path_inode_usage(self):

        inode_st = fsutil.get_path_inode_usage('/')

        self.assertGreaterEqual(inode_st['percent'], 0.0, '')
        self.assertLessEqual(inode_st['percent'], 1.0, '')
        self.assertLessEqual(inode_st['used'], inode_st['total'])

        total = inode_st['used'] + inode_st['available']
        self.assertEqual(inode_st['total'], total)

    def test_makedirs(self):

        fn = '/tmp/pykit-ut-fsutil-foo'
        fn_part = ('/tmp', 'pykit-ut-fsutil-foo')
        dd('fn_part:', fn_part)

        force_remove(fn)

        dd('file is not a dir')
        with open(fn, 'w') as f:
            f.write('a')
        self.assertRaises(OSError, fsutil.makedirs, fn)
        os.unlink(fn)

        dd('no error if dir exist')
        os.mkdir(fn)
        fsutil.makedirs(fn)
        os.rmdir(fn)

        dd('single part path should be created')
        fsutil.makedirs(fn)
        self.assertTrue(os.path.isdir(fn))
        os.rmdir(fn)

        dd('multi part path should be created')
        fsutil.makedirs(*fn_part)
        self.assertTrue(os.path.isdir(fn), 'multi part path should be created')
        os.rmdir(fn)

        dd('default mode')
        fsutil.makedirs(fn)
        self.assertEqual(0755, get_mode(fn))
        os.rmdir(fn)

        dd('specify mode')
        fsutil.makedirs(fn, mode=0700)
        self.assertEqual(0700, get_mode(fn))
        os.rmdir(fn)

        dd('specify uid/gid, to change uid, you need root privilege')
        fsutil.makedirs(fn, uid=1, gid=1)

    def test_get_sub_dirs(self):
        fsutil.makedirs('test_dir/sub_dir1')
        fsutil.makedirs('test_dir/sub_dir2')
        fsutil.write_file('test_dir/test_file', 'foo')

        sub_dirs = fsutil.get_sub_dirs('test_dir')
        self.assertListEqual(['sub_dir1', 'sub_dir2'], sub_dirs)

        fsutil.remove('test_dir')

    def test_makedirs_with_config(self):

        fn = '/tmp/pykit-ut-fsutil-foo'
        force_remove(fn)

        rc, out, err = proc.shell_script(pyt + ' ' + this_base + '/makedirs_with_config.py ' + fn,
                                         env=dict(PYTHONPATH=this_base + ':' + os.environ.get('PYTHONPATH'))
                                         )

        dd('run makedirs_with_config.py: ', rc, out, err)

        self.assertEqual(0, rc, 'normal exit')
        self.assertEqual('2,3', out, 'uid,gid is defined in test/pykitconfig.py')

    def test_read_write_file(self):

        fn = '/tmp/pykit-ut-rw-file'
        force_remove(fn)

        dd('write/read file')
        fsutil.write_file(fn, '123')
        self.assertEqual('123', fsutil.read_file(fn))

        dd('write/read 3MB file')
        cont = '123' * (1024**2)

        fsutil.write_file(fn, cont)
        self.assertEqual(cont, fsutil.read_file(fn))

        dd('write file with uid/gid')
        fsutil.write_file(fn, '1', uid=1, gid=1)
        stat = os.stat(fn)
        self.assertEqual(1, stat.st_uid)
        self.assertEqual(1, stat.st_gid)

        force_remove(fn)

    def test_write_file_with_config(self):

        fn = '/tmp/pykit-ut-fsutil-foo'
        force_remove(fn)

        rc, out, err = proc.shell_script(pyt + ' ' + this_base + '/write_with_config.py ' + fn,
                                         env=dict(PYTHONPATH=this_base + ':' + os.environ.get('PYTHONPATH'))
                                         )

        dd('run write_with_config.py: ', rc, out, err)

        self.assertEqual(0, rc, 'normal exit')
        self.assertEqual('2,3', out, 'uid,gid is defined in test/pykitconfig.py')

        force_remove(fn)

    def test_write_file_atomically(self):

        fn = '/tmp/pykit-ut-fsutil-write-atomic'

        dd('atomically write file')

        cont_thread1 = 'cont_thread1'
        cont_thread2 = 'cont_thread2'

        os_fsync = os.fsync

        def _wait_fsync(fildes):
            time.sleep(3)

            os_fsync(fildes)

        os.fsync = _wait_fsync

        assert_ok = {'ok': True}

        def _write_wait(cont_write, cont_read, start_after, atomic):

            time.sleep(start_after)

            fsutil.write_file(fn, cont_write, atomic=atomic)

            if cont_read != fsutil.read_file(fn):
                assert_ok['ok'] = False

        force_remove(fn)
        # atomic=False
        #  time     file    thread1     thread2
        #   0      cont_1   w_cont_1    sleep()
        #  1.5     cont_2   sleep()     w_cont_2
        #   3      cont_2   return      sleep()
        #  4.5     cont_2    None       return

        ths = []
        th = threadutil.start_daemon_thread(_write_wait,
                                            args=(cont_thread1, cont_thread2, 0, False))
        ths.append(th)

        th = threadutil.start_daemon_thread(_write_wait,
                                            args=(cont_thread2, cont_thread2, 1.5, False))
        ths.append(th)

        for th in ths:
            th.join()
        self.assertTrue(assert_ok['ok'])

        force_remove(fn)
        # atomic=True
        #  time     file    thread1     thread2
        #   0       None    w_cont_1    sleep()
        #  1.5      None    sleep()     w_cont_2
        #   3      cont_1   return      sleep()
        #  4.5     cont_2    None       return

        ths = []
        th = threadutil.start_daemon_thread(_write_wait,
                                            args=(cont_thread1, cont_thread1, 0, True))
        ths.append(th)

        th = threadutil.start_daemon_thread(_write_wait,
                                            args=(cont_thread2, cont_thread2, 1.5, True))
        ths.append(th)

        for th in ths:
            th.join()
        self.assertTrue(assert_ok['ok'])

        os.fsync = os_fsync
        force_remove(fn)

    def test_remove_normal_file(self):

        fn = '/tmp/pykit-ut-fsutil-remove-file-normal'
        force_remove(fn)

        fsutil.write_file(fn, '', atomic=True)
        self.assertTrue(os.path.isfile(fn))

        fsutil.remove(fn)
        self.assertFalse(os.path.exists(fn))

    def test_remove_link_file(self):

        src_fn = '/tmp/pykit-ut-fsutil-remove-file-normal'
        force_remove(src_fn)

        fsutil.write_file(src_fn, '', atomic=True)
        self.assertTrue(os.path.isfile(src_fn))

        link_fn = '/tmp/pykit-ut-fsutil-remove-file-link'
        force_remove(link_fn)

        os.link(src_fn, link_fn)
        self.assertTrue(os.path.isfile(link_fn))

        fsutil.remove(link_fn)
        self.assertFalse(os.path.exists(link_fn))

        symlink_fn = '/tmp/pykit-ut-fsutil-remove-file-symlink'
        force_remove(symlink_fn)

        os.symlink(src_fn, symlink_fn)
        self.assertTrue(os.path.islink(symlink_fn))

        fsutil.remove(symlink_fn)
        self.assertFalse(os.path.exists(symlink_fn))

        force_remove(src_fn)

    def test_remove_dir(self):

        dirname = '/tmp/pykit-ut-fsutil-remove-dir'

        fsutil.makedirs(dirname)
        self.assertTrue(os.path.isdir(dirname))

        for is_dir, file_path in (
                (False, ('normal_file',)),
                (True,  ('sub_dir',)),
                (False, ('sub_dir', 'sub_file1')),
                (False, ('sub_dir', 'sub_file2')),
                (True,  ('sub_empty_dir',)),
                (True,  ('sub_dir', 'sub_sub_dir')),
                (False, ('sub_dir', 'sub_sub_dir', 'sub_sub_file')),
        ):

            path = os.path.join(dirname, *file_path)

            if is_dir:
                fsutil.makedirs(path)
                self.assertTrue(os.path.isdir(path))
            else:
                fsutil.write_file(path, '')
                self.assertTrue(os.path.isfile(path))

        fsutil.remove(dirname)
        self.assertFalse(os.path.exists(dirname))

    def test_remove_dir_with_link(self):

        dirname = '/tmp/pykit-ut-fsutil-remove-dir'

        fsutil.makedirs(dirname)
        self.assertTrue(os.path.isdir(dirname))

        normal_file = 'normal_file'
        normal_path = os.path.join(dirname, normal_file)

        fsutil.write_file(normal_path, '')
        self.assertTrue(os.path.isfile(normal_path))

        hard_link = 'hard_link'
        hard_path = os.path.join(dirname, hard_link)

        os.link(normal_path, hard_path)
        self.assertTrue(os.path.isfile(hard_path))

        symbolic_link = 'symbolic_link'
        symbolic_path = os.path.join(dirname, symbolic_link)

        os.symlink(hard_path, symbolic_path)
        self.assertTrue(os.path.islink(symbolic_path))

        fsutil.remove(dirname)
        self.assertFalse(os.path.exists(dirname))

    def test_remove_error(self):

        dirname = '/tmp/pykit-ut-fsutil-remove-on-error'
        if os.path.isdir(dirname):
            fsutil.remove(dirname)

        # OSError
        self.assertRaises(os.error, fsutil.remove, dirname, False)

        # ignore errors
        fsutil.remove(dirname, ignore_errors=True)

        def assert_error(exp_func):
            def onerror(func, path, exc_info):
                self.assertEqual(func, exp_func)
            return onerror

        # on error
        fsutil.remove(dirname, onerror=assert_error(os.remove))

    def test_calc_checksums(self):

        M = 1024**2

        fn = '/tmp/pykit-ut-fsutil-calc_checksums'
        force_remove(fn)

        cases = (
            ('',
                {
                    'sha1':       True,
                    'md5':        True,
                    'crc32':      True,
                    'sha256':     True,
                    'block_size': M,
                    'io_limit':   M,
                },
                {
                    'sha1':       'da39a3ee5e6b4b0d3255bfef95601890afd80709',
                    'md5':        'd41d8cd98f00b204e9800998ecf8427e',
                    'crc32':      '00000000',
                    'sha256':     'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
                },
                None,
             ),

            ('',
                {
                    'md5':        True,
                    'crc32':      True,
                },
                {
                    'sha1':       None,
                    'md5':        'd41d8cd98f00b204e9800998ecf8427e',
                    'crc32':      '00000000',
                    'sha256':     None,
                },
                None,
             ),

            ('',
                {
                },
                {
                    'sha1':       None,
                    'md5':        None,
                    'crc32':      None,
                    'sha256':     None,
                },
                None,
             ),

            ('It  바로 とても 氣!',
                {
                    'sha1':       True,
                    'md5':        True,
                    'crc32':      True,
                    'sha256':     True,
                    'block_size': M,
                    'io_limit':   M,
                },
                {
                    'sha1':       'e22fa5446cb33d0c32221d89ee270dff23e32847',
                    'md5':        '9d245fca88360be492c715253d68ba6f',
                    'crc32':      '7c3becdb',
                    'sha256':     '7327753b12db5c0dd090ad802c1c8ff44ea4cb447f3091d43cab371bd7583d9a',
                },
                None,
             ),

            ('It  바로 とても 氣!',
                {
                    'sha1':       False,
                    'md5':        True,
                    'crc32':      True,
                    'sha256':     False,
                    'block_size': M,
                    'io_limit':   M,
                },
                {
                    'sha1':       None,
                    'md5':        '9d245fca88360be492c715253d68ba6f',
                    'crc32':      '7c3becdb',
                    'sha256':     None,
                },
                None,
             ),

            ('It  바로 とても 氣!',
                {
                    'sha1':       True,
                    'md5':        False,
                    'crc32':      False,
                    'sha256':     True,
                },
                {
                    'sha1':       'e22fa5446cb33d0c32221d89ee270dff23e32847',
                    'md5':        None,
                    'crc32':      None,
                    'sha256':     '7327753b12db5c0dd090ad802c1c8ff44ea4cb447f3091d43cab371bd7583d9a',
                },
                None,
             ),

            ('!' * M * 10,
                {
                    'sha1':       False,
                    'md5':        False,
                    'crc32':      False,
                    'sha256':     False,
                    'block_size': M,
                    'io_limit':   M,
                },
                {
                    'sha1':       None,
                    'md5':        None,
                    'crc32':      None,
                    'sha256':     None,
                },
                (0, 0.5),
             ),

            ('!' * M * 10,
                {
                    'sha1':       True,
                    'md5':        True,
                    'crc32':      True,
                    'sha256':     True,
                    'block_size': M * 10,
                    'io_limit':   M * 10,
                },
                {
                    'sha1':       'c5430d624c498024d0f3371670227a201e910054',
                    'md5':        '8f499b17375fc678c7256f3c0054db79',
                    'crc32':      'f0af209f',
                    'sha256':     'bd5263cc56b27fda9f86f41f6d2ec012eb60d757281003c363b88677c7dcc5e7',

                },
                (1, 1.5),
             ),

            ('!' * M * 10,
                {
                    'sha1':       True,
                    'block_size': M,
                    'io_limit':   M * 5,
                },
                None,
                (2, 2.5),
             ),

            ('!' * M * 10,
                {
                    'sha1':       True,
                    'block_size': M,
                    'io_limit': -1,
                },
                None,
                (0, 0.5),
             ),
        )

        for cont, args, exp_checksums, min_time in cases:

            force_remove(fn)
            fsutil.write_file(fn, cont)

            t0 = time.time()
            checksums = fsutil.calc_checksums(fn, **args)
            spend_time = time.time() - t0

            if exp_checksums is not None:
                for k in checksums:
                    self.assertEqual(exp_checksums[k], checksums[k],
                                     'except: {exp}, actuality: {act}'.format(
                        exp=exp_checksums[k],
                        act=checksums[k],
                    ))

            if min_time is not None:
                self.assertTrue(min_time[0] < spend_time < min_time[1])

        force_remove(fn)


def force_remove(fn):

    try:
        os.rmdir(fn)
    except:
        pass

    try:
        os.unlink(fn)
    except:
        pass


def get_mode(fn):
    mode = os.stat(fn).st_mode
    dd('mode read:', oct(mode))
    return mode & 0777



#!/usr/bin/env python2
# coding: utf-8

import os
import unittest

from pykit import argchecker


class TestChecker(unittest.TestCase):

    def test_check_arguments(self):
        cases = (
            (True, {'checker': {'type': 'bool'}}),
            (False, {'checker': {'type': 'bool'}}),

            (1, {'checker': {'type': 'integer'}}),
            (100, {'checker': {'type': 'integer'}}),

            (1, {'checker': {'type': 'float'}}),
            (50, {'checker': {'type': 'float'}}),
            (100.0, {'checker': {'type': 'float'}}),

            ('1', {'checker': {'type': 'string_number'}}),
            ('50', {'checker': {'type': 'string_number'}}),
            ('100.0', {'checker': {'type': 'string_number'}}),

            ('20', {'checker': {'type': 'string'}}),
            ('100.0', {'checker': {'type': 'string'}}),
            ('xxx', {'checker': {'type': 'string'}}),

            ([], {'checker': {'type': 'array'}}),
            ([1, 2], {'checker': {'type': 'array'}}),
            (['a', 'b'], {'checker': {'type': 'array'}}),

            ({'xx': {}}, {'xx': {'checker': {'type': 'dict'}}}),
            ({'xx': {'a': ''}}, {'xx': {'checker': {'type': 'dict'}}}),
        )

        for val, schema in cases:
            argchecker.check_arguments(val, schema)

    def test_required_arg(self):
        schema = {
            'x': {
                'checker': {'type': 'any'},
                'required': True
            }
        }
        with self.assertRaises(argchecker.LackArgumentError) as e:
            argchecker.check_arguments({}, schema)

        self.assertEqual(e.exception[0], 'x', "'x' is required")

    def test_undefined_arg(self):
        args = {'undefined': 0}
        schema = {'x': {'checker': {'type': 'any'}}}

        with self.assertRaises(argchecker.UnexpectedArgumentError) as e:
            argchecker.check_arguments(args, schema)

        self.assertEqual(e.exception[0], 'undefined')

    def test_float_arg(self):
        schema = {
            'num': {
                'required': True,
                'checker': {
                    'type': 'float',
                    'min': -100,
                    'max': 100,
                    'not': [0]
                }
            }
        }

        ex_cases = (
            ('1', argchecker.InvalidTypeError),
            (False, argchecker.InvalidTypeError),
            (-101, argchecker.OutOfRangeError),
            (101, argchecker.OutOfRangeError),
            (0, argchecker.InvalidValueError),
        )

        for val, err in ex_cases:
            self.assertRaises(err,
                              argchecker.check_arguments,
                              {'num': val},
                              schema)

        cases = (
            -100,
            -50,
            -10.88,
            50,
            99.99,
            100,
        )

        for val in cases:
            argchecker.check_arguments({'num': val}, schema)

    def test_integer_arg(self):
        schema = {
            'int': {
                'checker': {
                    'type': 'integer'
                }
            }
        }
        with self.assertRaises(argchecker.InvalidTypeError):
            argchecker.check_arguments({'int': 1.1}, schema)

        argchecker.check_arguments({'int': 100}, schema)

        long_num = 10 ** 21
        argchecker.check_arguments({'int': long_num}, schema)

    def test_string_number(self):
        schema = {
            'num': {
                'checker': {
                    'type': 'string_number',
                    'min': 0,
                    'max': 10,
                    'not': ['5'],
                }
            }
        }

        ex_cases = (
            (1, argchecker.InvalidTypeError),
            ('a', argchecker.InvalidTypeError),
            ('11', argchecker.OutOfRangeError),
            ('-1', argchecker.OutOfRangeError),
            ('5', argchecker.InvalidValueError),
        )

        for val, err in ex_cases:
            self.assertRaises(err,
                              argchecker.check_arguments,
                              {'num': val},
                              schema)

        cases = (
            '10',
            '0',
            '9.99',
            '4.30',
        )

        for val in cases:
            argchecker.check_arguments({'num': val}, schema)

    def test_string_and_string_number(self):
        schema = {
            'num': {
                'checker': [
                    {
                        'type': 'string_number'
                    },

                    {
                        'type': 'string',
                        'enum': ['auto']
                    }
                ]
            }
        }

        cases = (
            '123',
            'auto',
        )

        for val in cases:
            argchecker.check_arguments({'num': val}, schema)

    def test_fixed_len(self):
        str_schema = {
            'checker': {
                'type': 'string',
                'fixed_length': 3,
            },
        }

        array_schema = {
            'checker': {
                'type': 'array',
                'fixed_length': 3,
            },
        }

        dict_schema = {
            'xx': {
                'checker': {
                    'type': 'dict',
                    'fixed_length': 3,
                },
            }
        }

        cases = (
            ('abc', str_schema),
            (['a', 'b', 'abc'], array_schema),
            ({'xx': {'a': 'a', 'b': 'b', 'c': 'c'}}, dict_schema),
        )

        for v, s in cases:
            argchecker.check_arguments(v, s)

        ex_cases = (
            ('a', str_schema),
            ('ab', str_schema),
            ('abcd', str_schema),

            (['a'], array_schema),
            (['a', 'b'], array_schema),
            (['a', 'b', 'abc', 'abcd'], array_schema),

            ({'xx': {'a': ''}}, dict_schema),
            ({'xx': {'a': '', 'b': ''}}, dict_schema),
            ({'xx': {'a': '', 'b': '', 'c': '', 'd': ''}}, dict_schema),
        )

        for v, s in ex_cases:
            self.assertRaises(argchecker.InvalidLengthError,
                              argchecker.check_arguments,
                              v,
                              s)

    def test_min_max_len(self):
        str_schema = {
            'checker': {
                'type': 'string',
                'min_length': 2,
                'max_length': 4,
            }
        }

        array_schema = {
            'checker': {
                'type': 'array',
                'min_length': 2,
                'max_length': 4,
            }
        }

        dict_schema = {
            'xx': {
                'checker': {
                    'type': 'dict',
                    'min_length': 2,
                    'max_length': 4,
                }
            }
        }

        cases = (
            ('12', str_schema),
            ('123', str_schema),
            ('1234', str_schema),

            (['a', 'b'], array_schema),
            (['a', 'b', 'c'], array_schema),
            (['a', 'b', 'c', 'd'], array_schema),

            ({'xx': {'a': '', 'b': ''}}, dict_schema),
            ({'xx': {'a': '', 'b': '', 'c': ''}}, dict_schema),
            ({'xx': {'a': '', 'b': '', 'c': '', 'd': ''}}, dict_schema),
        )

        for v, s in cases:
            argchecker.check_arguments(v, s)

        ex_cases = (
            ('1', str_schema),
            ('12345', str_schema),

            (['a'], array_schema),
            (['a', 'b', 'c', 'd', 'e'], array_schema),

            ({'xx': {'a': ''}}, dict_schema),
            ({'xx': {'a': '', 'b': '', 'c': '', 'd': '', 'e': ''}}, dict_schema),
        )

        for v, s in ex_cases:
            self.assertRaises(argchecker.InvalidLengthError,
                              argchecker.check_arguments,
                              v,
                              s)

    def test_str_enum(self):
        enum = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                'Friday', 'Saturday', 'Sunday']
        schema = {
            'week_day': {
                'checker': {
                    'type': 'string',
                    'enum': enum,
                }
            }
        }

        ex_cases = (
            'xx',
            'foo',
            'bar',
            'today',
            'monday',
            'tUesday',
        )

        for c in ex_cases:
            self.assertRaises(argchecker.NotInEnumError,
                              argchecker.check_arguments,
                              {'week_day': c},
                              schema)

        for c in enum:
            argchecker.check_arguments({'week_day': c}, schema)

    def test_str_regexp(self):
        schema = {
            'alphabet_only': {
                'checker': {
                    'type': 'string',
                    'regexp': "^[a-z]+$"
                }
            }
        }

        with self.assertRaises(argchecker.PatternNotMatchError):
            argchecker.check_arguments({'alphabet_only': 'abcd123'}, schema)

        argchecker.check_arguments({'alphabet_only': 'abcd'}, schema)

    def test_array_arg(self):
        schema = {
            'numbers': {
                'checker': {
                    'type': 'array',
                    'element_checker': {
                        'type': 'integer'
                    }
                }
            }
        }
        with self.assertRaises(argchecker.InvalidTypeError):
            argchecker.check_arguments({'numbers': 123}, schema)

        with self.assertRaises(argchecker.InvalidTypeError):
            argchecker.check_arguments({'numbers': [1, 2, 'a']}, schema)

        argchecker.check_arguments({'numbers': []}, schema)
        argchecker.check_arguments({'numbers': [1, 2]}, schema)

    def test_dict_arg(self):
        schema = {
            'log_counter': {
                'checker': {
                    'type': 'dict',
                    'key_checker': {
                        'type': 'string',
                        'enum': ['info', 'error', 'warn']
                    },
                    'value_checker': {
                        'type': 'integer',
                        'min': 0
                    }
                }
            }
        }

        with self.assertRaises(argchecker.InvalidTypeError):
            argchecker.check_arguments({'log_counter': {1: 2}}, schema)

        with self.assertRaises(argchecker.NotInEnumError):
            argchecker.check_arguments({'log_counter': {'err': 1}}, schema)

        argchecker.check_arguments({'log_counter': {'error': 2, 'info': 1}},
                                   schema)

    def test_any_arg(self):
        schema = {
            'anything': {
                'checker': {'type': 'any'}
            }
        }
        for x in ([], {}, 0, ''):
            argchecker.check_arguments({'anything': x}, schema)

    def test_any_array(self):
        schema = {
            'anything': {
                'checker': {'type': 'array'}
            }
        }

        argchecker.check_arguments({'anything': [[], {}, 0, '']}, schema)

    def test_any_dict(self):
        schema = {
            'anything': {
                'checker': {'type': 'dict'}
            }
        }

        argchecker.check_arguments({'anything': {'1': 'a', 2: 'b', 3: [4]}}, schema)

    def test_nested_shema(self):
        sub_schema = {
            'nested_arg': {
                'checker': {'type': 'string'}
            }
        }
        schema = {
            'arg': {
                'checker': {
                    'type': 'dict',
                    'sub_schema': sub_schema
                }
            }
        }

        with self.assertRaises(argchecker.UnexpectedArgumentError) as e:
            argchecker.check_arguments({'arg': {'invalid_arg': '123'}}, schema)
        self.assertEqual(e.exception[0], 'invalid_arg')

        with self.assertRaises(argchecker.InvalidTypeError) as e:
            argchecker.check_arguments({'arg': {'nested_arg': 1}}, schema)
        self.assertEqual(e.exception[0], 'nested_arg')

        argchecker.check_arguments({'arg': {'nested_arg': '123'}}, schema)

    def test_checker_with_multi_types(self):
        schema = {
            'str_or_int': {
                'checker': [
                    {'type': 'string'},
                    {'type': 'integer'},
                ]
            }
        }

        cases = (
            'xx',
            123,
        )

        for val in cases:
            argchecker.check_arguments({'str_or_int': val}, schema)

        ex_cases = (
            [],
            (),
            {},
            0.1,
        )

        for val in ex_cases:
            self.assertRaises(argchecker.InvalidTypeError,
                              argchecker.check_arguments,
                              {'str_or_int': val},
                              schema)

    def test_read_schema(self):
        test_dir = os.path.dirname(os.path.abspath(__file__))
        test_schema_path = test_dir + '/schema_example.yaml'
        schema = argchecker.read_schema(test_schema_path)

        self.assertIs(type(schema), dict)

    def _validate_and_assert_error_msg(self, schema, expected_err_msg):
        with self.assertRaises(argchecker.InvalidSchemaError) as e:
            argchecker.validate_schema({}, schema)

        self.assertIn(expected_err_msg, str(e.exception))

    def test_empty_schema(self):
        for schema in (None, {}):
            self._validate_and_assert_error_msg(schema, "Empty Schema")

    def test_missing_required_keyword(self):
        schema = {'arg': {}}
        expected_err_msg = "'checker' is required for scope 'root'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

        schema = {'arg': {'checker': {}}}
        expected_err_msg = "'type' is required for scope 'checker'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_extra_keyword(self):
        schema = {'arg': {'checker': {'type': 'float'},
                          'extra_keyword': True}}

        expected_err_msg = "'extra_keyword' is not a valid keyword"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_invalid_keyword_type(self):
        schema = {'arg': {'checker': None}}
        expected_err_msg = "Expected types '['dict', 'array']' for 'checker', got 'None'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

        schema = {'arg': {'checker': {'type': 'float',
                                      'min': '0'}}}
        expected_err_msg = "Expected types '['float']' for 'min', got '0'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

        schema = {'arg': {'checker': {'type': 'float',
                                      'enum': [1, '2']}}}
        expected_err_msg = "Expected types '['float']' for 'enum', got '2'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_incorrect_keyword_scope(self):
        schema = {'arg': {'min': 0, 'checker': {'type': 'float'}}}
        expected_err_msg = "'min' is not in the correct scope "
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_invalid_scope_type(self):
        schema = {'arg': {'checker': {'type': 'array', 'min': 0}}}
        expected_err_msg = "'min' can only be used when the scope type is one of"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_nested_checker(self):
        value_checker = {
            'type': 'array',
            'element_checker': {
                'type': 'string',
                'enum': [1, 2, 3]
            }
        }
        schema = {
            'arg': {
                'checker': {
                    'type': 'dict',
                    'value_checker': value_checker
                }
            }
        }

        expected_err_msg = "Expected types '['string']' for 'enum', got '1'"
        self._validate_and_assert_error_msg(schema, expected_err_msg)

    def test_nested_schema(self):
        sub_schema = {
            'sub_arg': {
                'checker': {'type': 'string'}
            }
        }
        schema = {
            'arg': {
                'checker': {
                    'type': 'dict',
                    'sub_schema': sub_schema
                }
            }
        }

        argchecker.validate_schema({}, schema)

    def test_checker_multi_types(self):
        schema = {
            'arg': {
                'checker': [
                    {'type': 'string', 'enum': ['auto']},
                    {'type': 'integer'}
                ]
            }
        }

        argchecker.validate_schema({}, schema)
